from abjad.system.AbjadValueObject import AbjadValueObject


class Signature(AbjadValueObject):
    """
    Signature.

    Decorates expression-enabled methods.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_argument_list_callback',
        '_is_operator',
        '_markup_maker_callback',
        '_method_name',
        '_method_name_callback',
        '_string_template_callback',
        '_subscript',
        '_superscript',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        argument_list_callback=None,
        is_operator=None,
        markup_maker_callback=None,
        method_name=None,
        method_name_callback=None,
        string_template_callback=None,
        subscript=None,
        superscript=None,
        ):
        self._argument_list_callback = argument_list_callback
        self._is_operator = is_operator
        self._markup_maker_callback = markup_maker_callback
        self._method_name = method_name
        self._method_name_callback = method_name_callback
        self._string_template_callback = string_template_callback
        self._subscript = subscript
        self._superscript = superscript

    ### SPECIAL METHODS ###

    def __call__(self, method):
        """
        Calls signature decorator on ``method``.

        Returns ``method`` with metadata attached.
        """
        method.argument_list_callback = self.argument_list_callback
        method.is_operator = self.is_operator
        method.markup_maker_callback = self.markup_maker_callback
        method.method_name = self.method_name
        method.method_name_callback = self.method_name_callback
        method.string_template_callback = self.string_template_callback
        method.subscript = self.subscript
        method.superscript = self.superscript
        method.has_signature_decorator = True
        return method

    ### PUBLIC PROPERTIES ###

    @property
    def argument_list_callback(self):
        """
        Gets argument list callback.

        Returns string or none.
        """
        return self._argument_list_callback

    @property
    def is_operator(self):
        """
        Is true when method typesets like operator.

        Returns true, false or none.
        """
        return self._is_operator

    @property
    def markup_maker_callback(self):
        """
        Gets markup maker callback.

        Returns string or none.
        """
        return self._markup_maker_callback

    @property
    def method_name(self):
        """
        Gets method name.

        Returns string or none.
        """
        return self._method_name

    @property
    def method_name_callback(self):
        """
        Gets method name callback.

        Returns string or none.
        """
        return self._method_name_callback

    @property
    def string_template_callback(self):
        """
        Gets string template callback.

        Returns string or none.
        """
        return self._string_template_callback

    @property
    def subscript(self):
        """
        Gets subscript.

        Returns string or none.
        """
        return self._subscript

    @property
    def superscript(self):
        """
        Gets superscript.

        Returns string or none.
        """
        return self._superscript



import abjad
import ide
import os
import pathlib
import sys
import traceback


if __name__ == '__main__':

    try:
        from definition import maker
    except ImportError:
        traceback.print_exc()
        sys.exit(1)

    try:
        from __metadata__ import metadata as metadata
    except ImportError:
        traceback.print_exc()
        sys.exit(1)

    try:
        {previous_segment_metadata_import_statement}
    except ImportError:
        traceback.print_exc()
        sys.exit(1)

    try:
        segment_directory = pathlib.Path(os.path.realpath(__file__)).parent
        builds_directory = segment_directory.parent.parent / 'builds'
        builds_directory = ide.Path(builds_directory)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        with abjad.Timer() as timer:
            lilypond_file = maker.run(
                metadata=metadata,
                previous_metadata=previous_metadata,
                )
        segment_maker_runtime = int(timer.elapsed_time)
        count = segment_maker_runtime
        counter = abjad.String('second').pluralize(count)
        message = f'Segment-maker runtime {{count}} {{counter}} ...'
        print(message)
        segment_maker_runtime = (count, counter)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        segment = ide.Path(__file__).parent
        segment.write_metadata_py(maker.metadata)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        illustration_ly = segment('illustration.ly')
        result = abjad.persist(lilypond_file).as_ly(illustration_ly, strict=89)
        abjad_format_time = int(result[1])
        count = abjad_format_time
        counter = abjad.String('second').pluralize(count)
        message = f'Abjad format time {{count}} {{counter}} ...'
        print(message)
    except:
        traceback.print_exc()
        sys.exit(1)

    sys.exit(0)



def is_assignable_integer(argument):
    r"""
    Is true when ``argument`` is equivalent to an integer that can be written
    without recourse to ties.

    ..  container:: example

        >>> for n in range(0, 16 + 1):
        ...     print('%s\t%s' % (n, abjad.mathtools.is_assignable_integer(n)))
        ...
        0  False
        1  True
        2  True
        3  True
        4  True
        5  False
        6  True
        7  True
        8  True
        9  False
        10 False
        11 False
        12 True
        13 False
        14 True
        15 True
        16 True

    Returns true or false.
    """
    from abjad import mathtools
    if isinstance(argument, int):
        if 0 < argument:
            if '01' not in mathtools.integer_to_binary_string(argument):
                return True
    return False



import sys


def run_abjad():
    """
    Runs Abjad.

    Returns none.
    """
    from abjad import system

    try:
        file_name = sys.argv[1]
    except IndexError:
        file_name = ''

    commands = (
        "import abjad;",
        "from abjad import *;",
        "print(abjad_configuration.get_abjad_startup_string());",
        )
    commands = ' '.join(commands)

    command = r'''python -i {} -c "{}"'''
    command = command.format(file_name, commands)
    system.IOManager.spawn_subprocess(command)



"""
Abjad's LilyPond parser and supporting classes.
"""

from .Music import Music
from .ContextSpeccedMusic import ContextSpeccedMusic
from .GuileProxy import GuileProxy
from .LilyPondDuration import LilyPondDuration
from .LilyPondEvent import LilyPondEvent
from .LilyPondFraction import LilyPondFraction
from .LilyPondGrammarGenerator import LilyPondGrammarGenerator
from .LilyPondLexicalDefinition import LilyPondLexicalDefinition
from ._parse import _parse
from ._parse_debug import _parse_debug
from .LilyPondParser import LilyPondParser
from .LilyPondSyntacticalDefinition import LilyPondSyntacticalDefinition
from .ReducedLyParser import ReducedLyParser
from .SchemeParser import SchemeParser
from .SequentialMusic import SequentialMusic
from .SimultaneousMusic import SimultaneousMusic
from .SyntaxNode import SyntaxNode
from .parse_reduced_ly_syntax import parse_reduced_ly_syntax



def cumulative_sums(argument, start=0):
    """
    Gets cumulative sums of ``argument``.

    ..  container:: example

        >>> abjad.mathtools.cumulative_sums([1, 2, 3, 4, 5, 6, 7, 8], start=0)
        [0, 1, 3, 6, 10, 15, 21, 28, 36]

        >>> abjad.mathtools.cumulative_sums([1, 2, 3, 4, 5, 6, 7, 8], start=None)
        [1, 3, 6, 10, 15, 21, 28, 36]

    Raises exception when ``argument`` is not iterable.

    Returns new object of ``argument`` type.
    """
    if start is None:
        result = []
    else:
        result = [start]
    for item in argument:
        if result:
            new = result[-1] + item
            result.append(new)
        else:
            result.append(item)
    return type(argument)(result)



"""
Tools for modeling high-level constructs in LilyPond's file format.
"""

from .Block import Block
from .ContextBlock import ContextBlock
from .DateTimeToken import DateTimeToken
from .LilyPondDimension import LilyPondDimension
from .LilyPondFile import LilyPondFile
from .LilyPondLanguageToken import LilyPondLanguageToken
from .LilyPondVersionToken import LilyPondVersionToken
from .PackageGitCommitToken import PackageGitCommitToken



import abc


class AbjadObject(object, metaclass=abc.ABCMeta):
    """
    Abstract base class from which many custom classes inherit.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __format__(self, format_specification=''):
        """
        Formats Abjad object.

        Set ``format_specification`` to ``''`` or ``'storage'``.
        Interprets ``''`` equal to ``'storage'``.

        Returns string.
        """
        import abjad
        if format_specification in ('', 'storage'):
            return abjad.StorageFormatManager(self).get_storage_format()
        return str(self)

    def __getstate__(self):
        """
        Gets state of Abjad object.

        Returns dictionary.
        """
        if hasattr(self, '__dict__') and hasattr(vars(self), 'copy'):
            state = vars(self).copy()
        else:
            state = {}
        for class_ in type(self).__mro__:
            for slot in getattr(class_, '__slots__', ()):
                try:
                    state[slot] = getattr(self, slot)
                except AttributeError:
                    pass
        return state

    def __repr__(self):
        """
        Gets interpreter representation of Abjad object.

        Returns string.
        """
        import abjad
        return abjad.StorageFormatManager(self).get_repr_format()

    def __setstate__(self, state):
        """
        Sets state of Abjad object.

        Returns none.
        """
        for key, value in state.items():
            setattr(self, key, value)

    ### PRIVATE METHODS ###

    def _debug(self, value, annotation=None, blank=False):
        if annotation is None:
            print('debug: {!r}'.format(value))
        else:
            print('debug ({}): {!r}'.format(annotation, value))
        if blank:
            print()

    def _debug_values(self, values, annotation=None, blank=True):
        if values:
            for value in values:
                self._debug(value, annotation=annotation)
            if blank:
                print()
        else:
            self._debug(repr(values), annotation=annotation)
            if blank:
                print()

    def _get_format_specification(self):
        import abjad
        return abjad.FormatSpecification(client=self)



"""
Enumerations.
"""

import uqbar.enums


class Comparison(uqbar.enums.StrictEnumeration):
    """
    Enumeration of amount comparisons.
    """
    Less = -1
    Exact = 0
    More = 1

    def __format__(self, format_spec):
        return repr(self)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class HorizontalAlignment(uqbar.enums.StrictEnumeration):
    """
    Enumeration of horizontal alignments.
    """
    Left = -1
    Both = 0
    Right = 1

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class VerticalAlignment(uqbar.enums.StrictEnumeration):
    """
    Enumeration of vertical alignments.
    """
    Down = -1
    Center = 0
    Up = 1

    def __format__(self, format_spec):
        """
        Formats vertical alignment as ``storage`` or ``lilypond``.
        """
        if format_spec == 'lilypond':
            return {
                self.Down: '_',
                self.Up: '^',
                self.Center: '-',
            }[self]
        return repr(self)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_expr(cls, expr):
        lilypond_symbols = {
            '^': cls.Up,
            '-': cls.Center,
            '_': cls.Down,
        }
        result = lilypond_symbols.get(expr)
        if result is not None:
            return result
        return super().from_expr(expr)


Both = HorizontalAlignment.Both
Center = VerticalAlignment.Center
Down = VerticalAlignment.Down
Exact = Comparison.Exact
Left = HorizontalAlignment.Left
Less = Comparison.Less
More = Comparison.More
Right = HorizontalAlignment.Right
Up = VerticalAlignment.Up


__all__ = [
    'Both',
    'Center',
    'Comparison',
    'Down',
    'Exact',
    'HorizontalAlignment',
    'Left',
    'Less',
    'More',
    'Right',
    'Up',
    'VerticalAlignment',
    ]



import collections
import typing
from abjad import exceptions
from abjad.indicators.TimeSignature import TimeSignature
from abjad.markups import Markup
from abjad.pitch.NamedPitch import NamedPitch
from abjad.pitch.PitchSet import PitchSet
from abjad.spanners.Spanner import Spanner
from abjad.system.AbjadObject import AbjadObject
from abjad.system.LilyPondFormatManager import LilyPondFormatManager
from abjad.system.WellformednessManager import WellformednessManager
from abjad.system.Wrapper import Wrapper
from abjad.timespans.Timespan import Timespan
from abjad.top.inspect import inspect
from abjad.top.iterate import iterate
from abjad.utilities.Duration import Duration
from .AfterGraceContainer import AfterGraceContainer
from .Chord import Chord
from .Component import Component
from .Container import Container
from .GraceContainer import GraceContainer
from .Leaf import Leaf
from .Lineage import Lineage
from .LogicalTie import LogicalTie
from .Note import Note
from .Parentage import Parentage
from .Selection import Selection
from .Staff import Staff
from .Tuplet import Tuplet
from .VerticalMoment import VerticalMoment


class Inspection(AbjadObject):
    """
    Inspection.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.inspect(staff)
        Inspection(client=Staff("c'4 e'4 d'4 f'4"))

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Collaborators'

    __slots__ = (
        '_client',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        client: typing.Union[Component, typing.Iterable[Component]] = None,
        ) -> None:
        assert not isinstance(client, str), repr(client)
        prototype = (Component, collections.Iterable, type(None))
        if not isinstance(client, prototype):
            message = 'must be component, nonstring iterable or none:'
            message += f' (not {client!r}).'
            raise TypeError(message)
        self._client = client

    ### PUBLIC PROPERTIES ###

    @property
    def client(self) -> typing.Union[
        Component, typing.Iterable[Component], None
        ]:
        r"""
        Gets client of inspection.

        ..  container:: example

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Voice("c'8 d'8 e'8 f'8"))
            >>> staff.append(abjad.Voice("g'8 a'8 b'8 c''8"))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \new Voice
                    {
                        c'8
                        d'8
                        e'8
                        f'8
                    }
                    \new Voice
                    {
                        g'8
                        a'8
                        b'8
                        c''8
                    }
                }

            >>> abjad.inspect(staff).client
            <Staff{2}>

        """
        return self._client

    ### PUBLIC METHODS ###

    def after_grace_container(self) -> typing.Optional[
        AfterGraceContainer]:
        r"""
        Gets after grace containers attached to leaf.

        ..  container:: example

            Get after grace container attached to note:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> note = abjad.Note("ds'16")
            >>> container = abjad.AfterGraceContainer([note])
            >>> abjad.attach(container, staff[1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    \afterGrace
                    d'8
                    {
                        ds'16
                    }
                    e'8
                    f'8
                }

            >>> abjad.inspect(staff[1]).after_grace_container()
            AfterGraceContainer("ds'16")

        """
        return getattr(self.client, '_after_grace_container', None)

    def annotation(
        self,
        annotation,
        default=None,
        unwrap=True,
        ) -> typing.Any:
        r"""
        Gets annotation.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 e' e' f'")
            >>> abjad.annotate(staff[0], 'default_instrument', abjad.Cello())
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    e'4
                    e'4
                    f'4
                }

            >>> string = 'default_instrument'
            >>> abjad.inspect(staff[0]).annotation(string)
            Cello()

            >>> abjad.inspect(staff[1]).annotation(string) is None
            True

            >>> abjad.inspect(staff[2]).annotation(string) is None
            True

            >>> abjad.inspect(staff[3]).annotation(string) is None
            True

            Returns default when no annotation is found:

            >>> abjad.inspect(staff[3]).annotation(string, abjad.Violin())
            Violin()

        ..  container:: example

            REGRESSION: annotation is not picked up as effective indicator:

            >>> prototype = abjad.Instrument
            >>> abjad.inspect(staff[0]).effective(prototype) is None
            True

            >>> abjad.inspect(staff[1]).effective(prototype) is None
            True

            >>> abjad.inspect(staff[2]).effective(prototype) is None
            True

            >>> abjad.inspect(staff[3]).effective(prototype) is None
            True

        """
        assert isinstance(annotation, str), repr(annotation)
        for wrapper in self.annotation_wrappers():
            if wrapper.annotation == annotation:
                if unwrap is True:
                    return wrapper.indicator
                else:
                    return wrapper
        return default

    def annotation_wrappers(self) -> typing.List[Wrapper]:
        r"""
        Gets annotation wrappers.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 e' e' f'")
            >>> abjad.annotate(staff[0], 'default_instrument', abjad.Cello())
            >>> abjad.annotate(staff[0], 'default_clef', abjad.Clef('tenor'))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    e'4
                    e'4
                    f'4
                }

            >>> for wrapper in abjad.inspect(staff[0]).annotation_wrappers():
            ...     abjad.f(wrapper)
            ...
            abjad.Wrapper(
                annotation='default_instrument',
                indicator=abjad.Cello(
                    name='cello',
                    short_name='vc.',
                    markup=abjad.Markup(
                        contents=['Cello'],
                        ),
                    short_markup=abjad.Markup(
                        contents=['Vc.'],
                        ),
                    allowable_clefs=('bass', 'tenor', 'treble'),
                    context='Staff',
                    default_tuning=abjad.Tuning(
                        pitches=abjad.PitchSegment(
                            (
                                abjad.NamedPitch('c,'),
                                abjad.NamedPitch('g,'),
                                abjad.NamedPitch('d'),
                                abjad.NamedPitch('a'),
                                ),
                            item_class=abjad.NamedPitch,
                            ),
                        ),
                    middle_c_sounding_pitch=abjad.NamedPitch("c'"),
                    pitch_range=abjad.PitchRange('[C2, G5]'),
                    primary=True,
                    ),
                tag=abjad.Tag(),
                )
            abjad.Wrapper(
                annotation='default_clef',
                indicator=abjad.Clef('tenor'),
                tag=abjad.Tag(),
                )

        """
        result = []
        for wrapper in getattr(self.client, '_wrappers', []):
            if wrapper.annotation:
                result.append(wrapper)
        return result

    def badly_formed_components(self) -> typing.List[Component]:
        r"""
        Gets badly formed components.
        """
        manager = WellformednessManager()
        violators: typing.List[Component] = []
        for violators_, total, check_name in manager(self.client):
            violators.extend(violators_)
        return violators

    def contents(self, include_self=True) -> typing.Optional[Selection]:
        r"""
        Gets contents.

        ..  container:: example

            >>> staff = abjad.Staff(r"\times 2/3 { c'8 d'8 e'8 } f'4")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 2/3 {
                        c'8
                        d'8
                        e'8
                    }
                    f'4
                }

            >>> for component in abjad.inspect(staff).contents():
            ...     component
            ...
            <Staff{2}>
            Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            Note("f'4")

            >>> for component in abjad.inspect(staff).contents(
            ...     include_self=False,
            ...     ):
            ...     component
            ...
            Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            Note("f'4")

        """
        if not isinstance(self.client, Component):
            raise Exception('can only get contents of component.')
        return self.client._get_contents(include_self=include_self)

    def descendants(self, include_self=True) -> Selection:
        r"""
        Gets descendants.

        ..  container:: example

            >>> staff = abjad.Staff(r"\times 2/3 { c'8 d'8 e'8 } f'4")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 2/3 {
                        c'8
                        d'8
                        e'8
                    }
                    f'4
                }

            >>> for component in abjad.inspect(staff).descendants():
            ...     component
            ...
            <Staff{2}>
            Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            Note("c'8")
            Note("d'8")
            Note("e'8")
            Note("f'4")

            >>> for component in abjad.inspect(staff).descendants(
            ...     include_self=False,
            ...     ):
            ...     component
            ...
            Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            Note("c'8")
            Note("d'8")
            Note("e'8")
            Note("f'4")

            >>> for component in abjad.inspect(staff[:1]).descendants(
            ...     include_self=False,
            ...     ):
            ...     component
            ...
            Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            Note("c'8")
            Note("d'8")
            Note("e'8")

        """
        if isinstance(self.client, Component):
            return self.client._get_descendants(include_self=include_self)
        descendants: typing.List[Component] = []
        assert isinstance(self.client, Selection)
        for argument in self.client:
            descendants_ = Inspection(argument).descendants()
            for descendant_ in descendants_:
                if descendant_ not in descendants:
                    descendants.append(descendant_)
        result = Selection(descendants)
        return result

    def duration(self, in_seconds=False) -> Duration:
        r"""
        Gets duration.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }

            >>> selection = staff[:3]
            >>> abjad.inspect(selection).duration()
            Duration(3, 4)

        """
        if isinstance(self.client, Component):
            return self.client._get_duration(in_seconds=in_seconds)
        assert isinstance(self.client, collections.Iterable), repr(self.client)
        durations = [
            Inspection(_).duration(in_seconds=in_seconds)
            for _ in self.client
            ]
        return Duration(sum(durations))

    def effective(
        self,
        prototype=None,
        unwrap=True,
        n=0,
        default=None,
        ) -> typing.Any:
        r"""
        Gets effective indicator.

        ..  container:: example

            Gets effective clef:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Clef('alto'), staff[0])
            >>> abjad.attach(abjad.AcciaccaturaContainer("fs'16"), staff[-1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    \acciaccatura {
                        fs'16
                    }
                    f'4
                }

            >>> for component in abjad.iterate(staff).components():
            ...     clef = abjad.inspect(component).effective(abjad.Clef)
            ...     print(component, clef)
            ...
            Staff("c'4 d'4 e'4 f'4") Clef('alto')
            c'4 Clef('alto')
            d'4 Clef('alto')
            e'4 Clef('alto')
            fs'16 Clef('alto')
            f'4 Clef('alto')

        ..  container:: example

            Arbitrary objects (like strings) can be contexted:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> abjad.attach('color', staff[1], context='Staff')
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }

            >>> for component in abjad.iterate(staff).components():
            ...     string = abjad.inspect(component).effective(str)
            ...     print(component, repr(string))
            ...
            Staff("c'8 d'8 e'8 f'8") None
            c'8 None
            d'8 'color'
            e'8 'color'
            f'8 'color'

        ..  container:: example

            Scans forwards or backwards when ``n`` is set: 

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8 g'8")
            >>> abjad.attach('red', staff[0], context='Staff')
            >>> abjad.attach('blue', staff[2], context='Staff')
            >>> abjad.attach('yellow', staff[4], context='Staff')
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    d'8
                    e'8
                    f'8
                    g'8
                }
                
            >>> for n in (-1, 0, 1):
            ...     color = abjad.inspect(staff[0]).effective(str, n=n)
            ...     print(n, repr(color))
            ...
            -1 None
            0 'red'
            1 'blue'

            >>> for n in (-1, 0, 1):
            ...     color = abjad.inspect(staff[1]).effective(str, n=n)
            ...     print(n, repr(color))
            ...
            -1 None
            0 'red'
            1 'blue'

            >>> for n in (-1, 0, 1):
            ...     color = abjad.inspect(staff[2]).effective(str, n=n)
            ...     print(n, repr(color))
            ...
            -1 'red'
            0 'blue'
            1 'yellow'

            >>> for n in (-1, 0, 1):
            ...     color = abjad.inspect(staff[3]).effective(str, n=n)
            ...     print(n, repr(color))
            ...
            -1 'red'
            0 'blue'
            1 'yellow'

            >>> for n in (-1, 0, 1):
            ...     color = abjad.inspect(staff[4]).effective(str, n=n)
            ...     print(n, repr(color))
            ...
            -1 'blue'
            0 'yellow'
            1 None

        ..  container:: example

            Synthetic offsets works this way:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> abjad.attach(
            ...     'red',
            ...     staff[-1],
            ...     context='Staff',
            ...     synthetic_offset=-1,
            ...     )
            >>> abjad.attach('blue', staff[0], context='Staff')
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }

            Entire staff is effectively blue:

            >>> abjad.inspect(staff).effective(str)
            'blue'

            The (synthetic) offset just prior to (start of) staff is red:

            >>> abjad.inspect(staff).effective(str, n=-1)
            'red'

        ..  container:: example

            Gets effective time signature:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> leaves = abjad.select(staff).leaves()
            >>> abjad.attach(abjad.TimeSignature((3, 8)), leaves[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \time 3/8
                    c'4
                    d'4
                    e'4
                    f'4
                }

            >>> prototype = abjad.TimeSignature
            >>> for component in abjad.iterate(staff).components():
            ...     inspection = abjad.inspect(component)
            ...     time_signature = inspection.effective(prototype)
            ...     print(component, time_signature)
            ...
            Staff("c'4 d'4 e'4 f'4") 3/8
            c'4 3/8
            d'4 3/8
            e'4 3/8
            f'4 3/8

        """
        if not isinstance(self.client, Component):
            raise Exception('can only get effective on components.')
        result = self.client._get_effective(
            prototype=prototype,
            unwrap=unwrap,
            n=n,
            )
        if result is None:
            result = default
        return result

    def effective_staff(self) -> typing.Optional[Staff]:
        """
        Gets effective staff.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get effective staff on components.')
        return self.client._get_effective_staff()

    def effective_wrapper(
        self,
        prototype=None,
        n=0,
        ) -> typing.Optional[Wrapper]:
        r"""
        Gets effective wrapper.

        ..  container:: example

            Gets effective clef wrapper:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Clef('alto'), staff[0])
            >>> abjad.attach(abjad.AcciaccaturaContainer("fs'16"), staff[-1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    \acciaccatura {
                        fs'16
                    }
                    f'4
                }

            >>> for component in abjad.iterate(staff).components():
            ...     inspection = abjad.inspect(component)
            ...     wrapper = inspection.effective_wrapper(abjad.Clef)
            ...     print(component, wrapper)
            ...
            Staff("c'4 d'4 e'4 f'4") Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())
            c'4 Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())
            d'4 Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())
            e'4 Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())
            fs'16 Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())
            f'4 Wrapper(context='Staff', indicator=Clef('alto'), tag=Tag())

        """
        return self.effective(prototype=prototype, n=n, unwrap=False)

    def grace_container(self) -> typing.Optional[GraceContainer]:
        r"""
        Gets grace container attached to leaf.

        ..  container:: example

            Get acciaccatura container attached to note:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> note = abjad.Note("cs'16")
            >>> container = abjad.AcciaccaturaContainer([note])
            >>> abjad.attach(container, staff[1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    \acciaccatura {
                        cs'16
                    }
                    d'8
                    e'8
                    f'8
                }

            >>> abjad.inspect(staff[1]).grace_container()
            AcciaccaturaContainer("cs'16")

        """
        if not isinstance(self.client, Leaf):
            raise Exception('can only get grace container on leaf.')
        return self.client._grace_container

    def has_effective_indicator(self, prototype=None) -> bool:
        """
        Is true when client has effective indicator.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get effective indicator on component.')
        return self.client._has_effective_indicator(prototype=prototype)

    def has_indicator(self, prototype=None) -> bool:
        """
        Is true when client has one or more indicators.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get indicator on component.')
        return self.client._has_indicator(prototype=prototype)

    def has_spanner(self, prototype=None) -> bool:
        """
        Is true when client has one or more spanners.
        """
        if not isinstance(self.client, Leaf):
            raise Exception('can only get spanner on leaf.')
        return self.client._has_spanner(prototype=prototype)

    def indicator(
        self,
        prototype=None,
        default=None,
        unwrap=True,
        ) -> typing.Any:
        """
        Gets indicator.

        Raises exception when more than one indicator of ``prototype`` attach
        to client.

        Returns default when no indicator of ``prototype`` attaches to client.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get indicator on component.')
        indicators = self.client._get_indicators(
            prototype=prototype,
            unwrap=unwrap,
            )
        if not indicators:
            return default
        elif len(indicators) == 1:
            return list(indicators)[0]
        else:
            raise Exception('multiple indicators attached to client.')

    def indicators(self, prototype=None, unwrap=True) -> typing.List:
        r"""
        Get indicators.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Articulation('^'), staff[0])
            >>> abjad.attach(abjad.Articulation('^'), staff[1])
            >>> abjad.attach(abjad.Articulation('^'), staff[2])
            >>> abjad.attach(abjad.Articulation('^'), staff[3])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    -\marcato
                    d'4
                    -\marcato
                    e'4
                    -\marcato
                    f'4
                    -\marcato
                }

            >>> abjad.inspect(staff).indicators(abjad.Articulation)
            []

            >>> abjad.inspect(staff[0]).indicators(abjad.Articulation)
            [Articulation('^')]

        """
        # TODO: extend to any non-none client
        if not isinstance(self.client, Component):
            raise Exception('can only get indicators on component.')
        result = self.client._get_indicators(
            prototype=prototype,
            unwrap=unwrap,
            )
        return list(result)

    def is_bar_line_crossing(self) -> bool:
        r"""
        Is true when client crosses bar line.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d'4 e'4")
            >>> time_signature = abjad.TimeSignature((3, 8))
            >>> abjad.attach(time_signature, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \time 3/8
                    c'4
                    d'4
                    e'4
                }

            >>> for note in staff:
            ...     result = abjad.inspect(note).is_bar_line_crossing()
            ...     print(note, result)
            ...
            c'4 False
            d'4 True
            e'4 False

        """
        if not isinstance(self.client, Component):
            raise Exception('can only get indicator on component.')
        time_signature = self.client._get_effective(TimeSignature)
        if time_signature is None:
            time_signature_duration = Duration(4, 4)
        else:
            time_signature_duration = time_signature.duration
        partial = getattr(time_signature, 'partial', 0)
        partial = partial or 0
        start_offset = Inspection(self.client).timespan().start_offset
        shifted_start = start_offset - partial
        shifted_start %= time_signature_duration
        stop_offset = self.client._get_duration() + shifted_start
        if time_signature_duration < stop_offset:
            return True
        return False

    def is_grace_note(self) -> bool:
        """
        Is true when client is grace note.
        """
        if not isinstance(self.client, Leaf):
            return False
        prototype = (AfterGraceContainer, GraceContainer)
        for component in Inspection(self.client).parentage():
            if isinstance(component, prototype):
                return True
        return False

    def is_well_formed(
        self,
        check_beamed_long_notes=True,
        check_discontiguous_spanners=True,
        check_duplicate_ids=True,
        check_empty_containers=True,
        check_misdurated_measures=True,
        check_misfilled_measures=True,
        check_mismatched_enchained_hairpins=True,
        check_mispitched_ties=True,
        check_misrepresented_flags=True,
        check_missing_parents=True,
        check_nested_measures=True,
        check_notes_on_wrong_clef=True,
        check_out_of_range_notes=True,
        check_overlapping_beams=True,
        check_overlapping_glissandi=True,
        check_overlapping_hairpins=True,
        check_overlapping_octavation_spanners=True,
        check_overlapping_ties=True,
        check_overlapping_trill_spanners=True,
        check_tied_rests=True,
        ) -> bool:
        """
        Is true when client is well-formed.
        """
        manager = WellformednessManager()
        for violators, total, check_name in manager(self.client):
            if eval(check_name) is not True:
                continue
            if violators:
                return False
        return True

    def leaf(self, n=0) -> typing.Optional[Leaf]:
        r"""
        Gets leaf ``n``.

        ..  container:: example

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Voice("c'8 d'8 e'8 f'8"))
            >>> staff.append(abjad.Voice("g'8 a'8 b'8 c''8"))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \new Voice
                    {
                        c'8
                        d'8
                        e'8
                        f'8
                    }
                    \new Voice
                    {
                        g'8
                        a'8
                        b'8
                        c''8
                    }
                }

        ..  container:: example

            Gets leaf ``n`` **from** client when client is a leaf.

            With positive indices:

            >>> first_leaf = staff[0][0]
            >>> first_leaf
            Note("c'8")

            >>> for n in range(8):
            ...     leaf = abjad.inspect(first_leaf).leaf(n)
            ...     print(n, leaf)
            ...
            0 c'8
            1 d'8
            2 e'8
            3 f'8
            4 None
            5 None
            6 None
            7 None

            With negative indices:

            >>> last_leaf = staff[0][-1]
            >>> last_leaf
            Note("f'8")

            >>> for n in range(0, -8, -1):
            ...     leaf = abjad.inspect(last_leaf).leaf(n)
            ...     print(n, leaf)
            ...
            0 f'8
            -1 e'8
            -2 d'8
            -3 c'8
            -4 None
            -5 None
            -6 None
            -7 None

        ..  container:: example

            Gets leaf ``n`` **in** client when client is a container.

            With positive indices:

            >>> first_voice = staff[0]
            >>> first_voice
            Voice("c'8 d'8 e'8 f'8")

            >>> for n in range(8):
            ...     leaf = abjad.inspect(first_voice).leaf(n)
            ...     print(n, leaf)
            ...
            0 c'8
            1 d'8
            2 e'8
            3 f'8
            4 None
            5 None
            6 None
            7 None

            With negative indices:

            >>> first_voice = staff[0]
            >>> first_voice
            Voice("c'8 d'8 e'8 f'8")

            >>> for n in range(-1, -9, -1):
            ...     leaf = abjad.inspect(first_voice).leaf(n)
            ...     print(n, leaf)
            ...
            -1 f'8
            -2 e'8
            -3 d'8
            -4 c'8
            -5 None
            -6 None
            -7 None
            -8 None

        """
        if isinstance(self.client, Leaf):
            return self.client._get_leaf(n)
        if 0 <= n:
            reverse = False
        else:
            reverse = True
            n = abs(n) - 1
        leaves = iterate(self.client).leaves(reverse=reverse)
        for i, leaf in enumerate(leaves):
            if i == n:
                return leaf
        return None

    def lineage(self) -> Lineage:
        """
        Gets lineage.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get lineage on component.')
        return self.client._get_lineage()

    def logical_tie(self) -> LogicalTie:
        """
        Gets logical tie.
        """
        if not isinstance(self.client, Leaf):
            raise Exception('can only get logical tie on leaf.')
        return self.client._get_logical_tie()

    def markup(self, direction=None) -> typing.List[Markup]:
        """
        Gets markup.
        """
        # TODO: extend to any non-none client
        if not isinstance(self.client, Component):
            raise Exception('can only get markup on component.')
        result = self.client._get_markup(direction=direction)
        return list(result)

    def parentage(
        self,
        include_self=True,
        grace_notes=False,
        ) -> Parentage:
        r"""
        Gets parentage.

        .. container:: example

            Gets parentage without grace notes:

            >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
            >>> container = abjad.GraceContainer("c'16 d'16")
            >>> abjad.attach(container, voice[1])
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(voice)
                \new Voice
                {
                    c'4
                    \grace {
                        c'16
                        d'16
                    }
                    d'4
                    e'4
                    f'4
                }

            >>> abjad.inspect(container[0]).parentage()
            Parentage(component=Note("c'16"))

        .. container:: example

            Gets parentage with grace notes:

            >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
            >>> container = abjad.GraceContainer("c'16 d'16")
            >>> abjad.attach(container, voice[1])
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(voice)
                \new Voice
                {
                    c'4
                    \grace {
                        c'16
                        d'16
                    }
                    d'4
                    e'4
                    f'4
                }

            >>> agent = abjad.inspect(container[0])
            >>> parentage = agent.parentage(grace_notes=True)
            >>> for component in parentage:
            ...     component
            ...
            Note("c'16")
            GraceContainer("c'16 d'16")
            Note("d'4")
            Voice("c'4 d'4 e'4 f'4")

        """
        if not isinstance(self.client, Component):
            raise Exception('can only get parentage on component.')
        return self.client._get_parentage(
            include_self=include_self,
            grace_notes=grace_notes,
            )

    def pitches(self) -> typing.Optional[PitchSet]:
        """
        Gets pitches.
        """
        if not self.client:
            return None
        selection = Selection(self.client)
        return PitchSet.from_selection(selection)

    def report_modifications(self) -> str:
        r"""
        Reports modifications.

        ..  container:: example

            Reports container modifications:

            >>> container = abjad.Container("c'8 d'8 e'8 f'8")
            >>> abjad.override(container).note_head.color = 'red'
            >>> abjad.override(container).note_head.style = 'harmonic'
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    \override NoteHead.color = #red
                    \override NoteHead.style = #'harmonic
                    c'8
                    d'8
                    e'8
                    f'8
                    \revert NoteHead.color
                    \revert NoteHead.style
                }

            >>> report = abjad.inspect(container).report_modifications()
            >>> print(report)
            {
                \override NoteHead.color = #red
                \override NoteHead.style = #'harmonic
                %%% 4 components omitted %%%
                \revert NoteHead.color
                \revert NoteHead.style
            }

        ..  container:: example

            Reports leaf modifications:

            >>> container = abjad.Container("c'8 d'8 e'8 f'8")
            >>> abjad.attach(abjad.Clef('alto'), container[0])
            >>> abjad.override(container[0]).note_head.color = 'red'
            >>> abjad.override(container[0]).stem.color = 'red'
            >>> abjad.show(container) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(container)
                {
                    \once \override NoteHead.color = #red
                    \once \override Stem.color = #red
                    \clef "alto"
                    c'8
                    d'8
                    e'8
                    f'8
                }

            >>> report = abjad.inspect(container[0]).report_modifications()
            >>> print(report)
            slot absolute before:
            slot 1:
                grob overrides:
                    \once \override NoteHead.color = #red
                    \once \override Stem.color = #red
            slot 3:
                commands:
                    \clef "alto"
            slot 4:
                leaf body:
                    c'8
            slot 5:
            slot 7:
            slot absolute after:

        """
        if isinstance(self.client, Container):
            bundle = LilyPondFormatManager.bundle_format_contributions(
                self.client
                )
            result: typing.List[str] = []
            for slot in ('before', 'open brackets', 'opening'):
                lines = self.client._get_format_contributions_for_slot(
                    slot,
                    bundle,
                    )
                result.extend(lines)
            line = f'    %%% {len(self.client)} components omitted %%%'
            result.append(line)
            for slot in ('closing', 'close brackets', 'after'):
                lines = self.client._get_format_contributions_for_slot(
                    slot,
                    bundle,
                    )
                result.extend(lines)
            return '\n'.join(result)
        elif isinstance(self.client, Leaf):
            return self.client._report_format_contributions()
        else:
            return f'only defined for components: {self.client}.'

    def sounding_pitch(self) -> NamedPitch:
        r"""
        Gets sounding pitch.

        ..  container:: example

            >>> staff = abjad.Staff("d''8 e''8 f''8 g''8")
            >>> piccolo = abjad.Piccolo()
            >>> abjad.attach(piccolo, staff[0])
            >>> abjad.Instrument.transpose_from_sounding_pitch(staff)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    d'8
                    e'8
                    f'8
                    g'8
                }

        """
        if not isinstance(self.client, Note):
            raise Exception('can only get sounding pitch of note.')
        return self.client._get_sounding_pitch()

    # TODO: return PitchSet instead of list
    def sounding_pitches(self) -> typing.List[NamedPitch]:
        r"""
        Gets sounding pitches.

        ..  container:: example

            >>> staff = abjad.Staff("<c''' e'''>4 <d''' fs'''>4")
            >>> glockenspiel = abjad.Glockenspiel()
            >>> abjad.attach(glockenspiel, staff[0])
            >>> abjad.Instrument.transpose_from_sounding_pitch(staff)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    <c' e'>4
                    <d' fs'>4
                }

            >>> abjad.inspect(staff[0]).sounding_pitches()
            [NamedPitch("c'''"), NamedPitch("e'''")]

        Returns tuple.
        """
        # TODO: extend to any non-none client
        if not isinstance(self.client, Chord):
            raise Exception('can only get sounding pitches of chord.')
        result = self.client._get_sounding_pitches()
        return list(result)

    def spanner(
        self,
        prototype=None,
        default=None,
        ) -> typing.Any:
        """
        Gets spanner.

        Raises exception when more than one spanner of ``prototype`` attaches
        to client.

        Returns ``default`` when no spanner of ``prototype`` attaches to
        client.
        """
        # TODO: extend to any non-none client
        if not isinstance(self.client, (Leaf, Selection)):
            raise Exception('can only get spanner on leaf or selection.')
        spanners = self.client._get_spanners(prototype=prototype)
        assert isinstance(spanners, list), repr(spanners)
        if not spanners:
            return default
        elif len(spanners) == 1:
            return spanners[0]
        else:
            raise exceptions.ExtraSpannerError

    def spanners(self, prototype=None) -> typing.List[Spanner]:
        r"""
        Gets spanners.

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d' e' f'")
            >>> abjad.attach(abjad.Beam(), staff[:2])
            >>> abjad.attach(abjad.Beam(), staff[2:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    [
                    d'8
                    ]
                    e'8
                    [
                    f'8
                    ]
                }

            >>> abjad.inspect(staff).spanners()
            []

            >>> abjad.inspect(staff[0]).spanners()
            [Beam("c'8, d'8", durations=(), span_beam_count=1)]

            >>> beams = abjad.inspect(staff[:]).spanners()
            >>> beams = list(beams)
            >>> beams.sort()
            >>> beams
            [Beam("c'8, d'8", durations=(), span_beam_count=1), Beam("e'8, f'8", durations=(), span_beam_count=1)]

        """
        if isinstance(self.client, Container):
            return []
        if isinstance(self.client, (Leaf, Selection)):
            return self.client._get_spanners(prototype=prototype)
        assert isinstance(self.client, collections.Iterable), repr(self.client)
        known_ids: typing.List[int] = []
        result = []
        for item in self.client:
            for spanner in inspect(item).spanners(prototype=prototype):
                id_ = id(spanner)
                if id_ not in known_ids:
                    known_ids.append(id_)
                    result.append(spanner)
        return result

    def tabulate_wellformedness(
        self,
        allow_percussion_clef=None,
        check_beamed_long_notes=True,
        check_discontiguous_spanners=True,
        check_duplicate_ids=True,
        check_empty_containers=True,
        check_misdurated_measures=True,
        check_misfilled_measures=True,
        check_mismatched_enchained_hairpins=True,
        check_mispitched_ties=True,
        check_misrepresented_flags=True,
        check_missing_parents=True,
        check_nested_measures=True,
        check_notes_on_wrong_clef=True,
        check_out_of_range_notes=True,
        check_overlapping_beams=True,
        check_overlapping_glissandi=True,
        check_overlapping_hairpins=True,
        check_overlapping_octavation_spanners=True,
        check_overlapping_ties=True,
        check_overlapping_trill_spanners=True,
        check_tied_rests=True,
        ) -> str:
        r"""
        Tabulates well-formedness.
        """
        manager = WellformednessManager(
            allow_percussion_clef=allow_percussion_clef,
            )
        triples = manager(self.client)
        strings = []
        for violators, total, check_name in triples:
            if eval(check_name) is not True:
                continue
            violator_count = len(violators)
            check_name = check_name.replace('check_', '')
            check_name = check_name.replace('_', ' ')
            string = f'{violator_count} /\t{total} {check_name}'
            strings.append(string)
        return '\n'.join(strings)

    def timespan(self, in_seconds=False) -> Timespan:
        r"""
        Gets timespan.

        ..  container:: example

            Gets timespan of grace notes:

            >>> voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
            >>> grace_notes = [abjad.Note("c'16"), abjad.Note("d'16")]
            >>> container = abjad.GraceContainer(grace_notes)
            >>> abjad.attach(container, voice[1])
            >>> container = abjad.AfterGraceContainer("e'16 f'16")
            >>> abjad.attach(container, voice[1])
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(voice)
                \new Voice
                {
                    c'8
                    [
                    \grace {
                        c'16
                        d'16
                    }
                    \afterGrace
                    d'8
                    {
                        e'16
                        f'16
                    }
                    e'8
                    f'8
                    ]
                }

            >>> for leaf in abjad.iterate(voice).leaves():
            ...     timespan = abjad.inspect(leaf).timespan()
            ...     print(str(leaf) + ':')
            ...     abjad.f(timespan)
            ...
            c'8:
            abjad.Timespan(
                start_offset=abjad.Offset(0, 1),
                stop_offset=abjad.Offset(1, 8),
                )
            c'16:
            abjad.Timespan(
                start_offset=abjad.Offset(
                    (1, 8),
                    grace_displacement=abjad.Duration(-1, 8),
                    ),
                stop_offset=abjad.Offset(
                    (1, 8),
                    grace_displacement=abjad.Duration(-1, 16),
                    ),
                )
            d'16:
            abjad.Timespan(
                start_offset=abjad.Offset(
                    (1, 8),
                    grace_displacement=abjad.Duration(-1, 16),
                    ),
                stop_offset=abjad.Offset(1, 8),
                )
            d'8:
            abjad.Timespan(
                start_offset=abjad.Offset(1, 8),
                stop_offset=abjad.Offset(1, 4),
                )
            e'16:
            abjad.Timespan(
                start_offset=abjad.Offset(
                    (1, 4),
                    grace_displacement=abjad.Duration(-1, 8),
                    ),
                stop_offset=abjad.Offset(
                    (1, 4),
                    grace_displacement=abjad.Duration(-1, 16),
                    ),
                )
            f'16:
            abjad.Timespan(
                start_offset=abjad.Offset(
                    (1, 4),
                    grace_displacement=abjad.Duration(-1, 16),
                    ),
                stop_offset=abjad.Offset(1, 4),
                )
            e'8:
            abjad.Timespan(
                start_offset=abjad.Offset(1, 4),
                stop_offset=abjad.Offset(3, 8),
                )
            f'8:
            abjad.Timespan(
                start_offset=abjad.Offset(3, 8),
                stop_offset=abjad.Offset(1, 2),
                )

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d' e' f'")
            >>> abjad.attach(abjad.Beam(), staff[:2])
            >>> abjad.attach(abjad.Beam(), staff[2:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    [
                    d'8
                    ]
                    e'8
                    [
                    f'8
                    ]
                }

            >>> abjad.inspect(staff).timespan()
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(1, 2))

            >>> abjad.inspect(staff[0]).timespan()
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(1, 8))

            >>> abjad.inspect(staff[:3]).timespan()
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(3, 8))

        """
        if isinstance(self.client, Component):
            return self.client._get_timespan(in_seconds=in_seconds)
        assert isinstance(self.client, collections.Iterable), repr(self.client)
        remaining_items = []
        for i, item in enumerate(self.client):
            if i == 0:
                first_item = item
            else:
                remaining_items.append(item)
        timespan = Inspection(first_item).timespan(in_seconds=in_seconds)
        start_offset = timespan.start_offset
        stop_offset = timespan.stop_offset
        for item in remaining_items:
            timespan = Inspection(item).timespan(in_seconds=in_seconds)
            if timespan.start_offset < start_offset:
                start_offset = timespan.start_offset
            if stop_offset < timespan.stop_offset:
                stop_offset = timespan.stop_offset
        return Timespan(start_offset, stop_offset)

    def tuplet(self, n=0) -> typing.Optional[Tuplet]:
        r"""
        Gets tuplet ``n``.

        ..  container:: example

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Tuplet((2, 3), "c'8 d' e'"))
            >>> staff.append(abjad.Tuplet((2, 3), "d'8 e' f'"))
            >>> staff.append(abjad.Tuplet((2, 3), "e'8 f' g'"))
            >>> staff.append(abjad.Tuplet((2, 3), "f'8 g' a'"))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 2/3 {
                        c'8
                        d'8
                        e'8
                    }
                    \times 2/3 {
                        d'8
                        e'8
                        f'8
                    }
                    \times 2/3 {
                        e'8
                        f'8
                        g'8
                    }
                    \times 2/3 {
                        f'8
                        g'8
                        a'8
                    }
                }

        ..  container:: example

            >>> for n in range(4):
            ...     tuplet = abjad.inspect(staff).tuplet(n)
            ...     print(n, tuplet)
            ...
            0 Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            1 Tuplet(Multiplier(2, 3), "d'8 e'8 f'8")
            2 Tuplet(Multiplier(2, 3), "e'8 f'8 g'8")
            3 Tuplet(Multiplier(2, 3), "f'8 g'8 a'8")

            >>> for n in range(-1, -5, -1):
            ...     tuplet = abjad.inspect(staff).tuplet(n)
            ...     print(n, tuplet)
            ...
            -1 Tuplet(Multiplier(2, 3), "f'8 g'8 a'8")
            -2 Tuplet(Multiplier(2, 3), "e'8 f'8 g'8")
            -3 Tuplet(Multiplier(2, 3), "d'8 e'8 f'8")
            -4 Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")

        """
        if 0 <= n:
            reverse = False
        else:
            reverse = True
            n = abs(n) - 1
        tuplets = iterate(self.client).components(Tuplet, reverse=reverse)
        for i, tuplet in enumerate(tuplets):
            if i == n:
                return tuplet
        return None

    def vertical_moment(self, governor=None) -> VerticalMoment:
        r"""
        Gets vertical moment.

        ..  container:: example

            >>> score = abjad.Score()
            >>> tuplet = abjad.Tuplet((4, 3), "d''8 c''8 b'8")
            >>> score.append(abjad.Staff([tuplet]))
            >>> staff_group = abjad.StaffGroup(lilypond_type='PianoStaff')
            >>> staff_group.append(abjad.Staff("a'4 g'4"))
            >>> staff_group.append(abjad.Staff("f'8 e'8 d'8 c'8"))
            >>> clef = abjad.Clef('bass')
            >>> abjad.attach(clef, staff_group[1][0])
            >>> score.append(staff_group)
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(score)
                \new Score
                <<
                    \new Staff
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 4/3 {
                            d''8
                            c''8
                            b'8
                        }
                    }
                    \new PianoStaff
                    <<
                        \new Staff
                        {
                            a'4
                            g'4
                        }
                        \new Staff
                        {
                            \clef "bass"
                            f'8
                            e'8
                            d'8
                            c'8
                        }
                    >>
                >>

            >>> agent = abjad.inspect(staff_group[1][0])
            >>> moment = agent.vertical_moment(governor=staff_group)
            >>> moment.leaves
            Selection([Note("a'4"), Note("f'8")])

            >>> agent = abjad.inspect(staff_group[1][1])
            >>> moment = agent.vertical_moment(governor=staff_group)
            >>> moment.leaves
            Selection([Note("a'4"), Note("e'8")])

            >>> agent = abjad.inspect(staff_group[1][2])
            >>> moment = agent.vertical_moment(governor=staff_group)
            >>> moment.leaves
            Selection([Note("g'4"), Note("d'8")])

            >>> agent = abjad.inspect(staff_group[1][3])
            >>> moment = agent.vertical_moment(governor=staff_group)
            >>> moment.leaves
            Selection([Note("g'4"), Note("c'8")])

        """
        if not isinstance(self.client, Component):
            raise Exception('can only get vertical moment on component.')
        return self.client._get_vertical_moment(governor=governor)

    def vertical_moment_at(self, offset) -> VerticalMoment:
        """
        Gets vertical moment at ``offset``.
        """
        if not isinstance(self.client, Component):
            raise Exception('can only get vertical moment on component.')
        return self.client._get_vertical_moment_at(offset)

    def wrapper(self, prototype=None) -> typing.Optional[Wrapper]:
        r"""
        Gets wrapper.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Articulation('^'), staff[0])
            >>> abjad.attach(abjad.Articulation('^'), staff[1])
            >>> abjad.attach(abjad.Articulation('^'), staff[2])
            >>> abjad.attach(abjad.Articulation('^'), staff[3])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    -\marcato
                    d'4
                    -\marcato
                    e'4
                    -\marcato
                    f'4
                    -\marcato
                }

            >>> abjad.inspect(staff).wrapper(abjad.Articulation) is None
            True

            >>> abjad.inspect(staff[0]).wrapper(abjad.Articulation)
            Wrapper(indicator=Articulation('^'), tag=Tag())

        Raises exception when more than one indicator of ``prototype`` attach
        to client.
        """
        return self.indicator(prototype=prototype, unwrap=False)

    def wrappers(self, prototype=None) -> typing.Optional[
        typing.List[Wrapper]
        ]:
        r"""
        Gets wrappers.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Articulation('^'), staff[0])
            >>> abjad.attach(abjad.Articulation('^'), staff[1])
            >>> abjad.attach(abjad.Articulation('^'), staff[2])
            >>> abjad.attach(abjad.Articulation('^'), staff[3])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    -\marcato
                    d'4
                    -\marcato
                    e'4
                    -\marcato
                    f'4
                    -\marcato
                }

            >>> abjad.inspect(staff).wrappers(abjad.Articulation)
            []

            >>> abjad.inspect(staff[0]).wrappers(abjad.Articulation)
            [Wrapper(indicator=Articulation('^'), tag=Tag())]

        """
        return self.indicators(prototype=prototype, unwrap=False)



def deactivate(text, tag, skipped=False):
    r"""
    Deactivates ``tag`` in ``text``.

    ..  container:: example

        Writes (active) tag into LilyPond input:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup('Allegro').with_color('red')
        >>> abjad.attach(
        ...     markup,
        ...     staff[0],
        ...     tag='RED_MARKUP',
        ...     )

        >>> text = format(staff, 'lilypond')
        >>> text = abjad.LilyPondFormatManager.left_shift_tags(text)
        >>> print(text)
        \new Staff {
            c'4
            - \markup {     %! RED_MARKUP
                \with-color %! RED_MARKUP
                    #red    %! RED_MARKUP
                    Allegro %! RED_MARKUP
                }           %! RED_MARKUP
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        Deactivates tag:

        >>> text = format(staff, 'lilypond')
        >>> text, count = abjad.deactivate(text, 'RED_MARKUP')
        >>> print(text)
        \new Staff {
            c'4
        %%% - \markup {     %! RED_MARKUP
        %%%     \with-color %! RED_MARKUP
        %%%         #red    %! RED_MARKUP
        %%%         Allegro %! RED_MARKUP
        %%%     }           %! RED_MARKUP
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip('\n') for _ in text.split('\n')]
        >>> lilypond_file = abjad.LilyPondFile.new(lines)
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Activates tag again:

        >>> text, count = abjad.activate(text, 'RED_MARKUP')
        >>> print(text)
        \new Staff {
            c'4
            - \markup {     %! RED_MARKUP
                \with-color %! RED_MARKUP
                    #red    %! RED_MARKUP
                    Allegro %! RED_MARKUP
                }           %! RED_MARKUP
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip('\n') for _ in text.split('\n')]
        >>> lilypond_file = abjad.LilyPondFile.new(lines)
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Deactivates tag again:

        >>> text, count = abjad.deactivate(text, 'RED_MARKUP')
        >>> print(text)
        \new Staff {
            c'4
        %%% - \markup {     %! RED_MARKUP
        %%%     \with-color %! RED_MARKUP
        %%%         #red    %! RED_MARKUP
        %%%         Allegro %! RED_MARKUP
        %%%     }           %! RED_MARKUP
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip('\n') for _ in text.split('\n')]
        >>> lilypond_file = abjad.LilyPondFile.new(lines)
        >>> abjad.show(lilypond_file) # doctest: +SKIP

    Tags can toggle indefinitely.

    Returns text, count pair.

    Count gives number of deactivated tags.
    """
    import abjad
    assert isinstance(tag, str) or callable(tag), repr(tag)
    lines, count, skipped_count = [], 0, 0
    treated_last_line, last_index = False, None
    found_already_deactivated_on_last_line = False
    text_lines = text.split('\n')
    text_lines = [_ + '\n' for _ in text_lines[:-1]] + text_lines[-1:]
    lines = []
    for line in text_lines:
        first_nonwhitespace_index = len(line) - len(line.lstrip())
        index = first_nonwhitespace_index
        if not abjad.Line(line).match(tag):
            lines.append(line)
            treated_last_line, last_index = False, None
            found_already_deactivated_on_last_line = False
            continue
        if line[index] != '%':
            if last_index is None:
                last_index = index
            if ' %@%' in line:
                prefix = '%@% '
                line = line.replace(' %@%', '')
            else:
                prefix = '%%% '
            target = line[last_index-4:last_index]
            assert target == '    ', repr((line, target, index, tag))
            characters = list(line)
            characters[last_index-4:last_index] = list(prefix)
            line = ''.join(characters)
            if not treated_last_line:
                count += 1
            treated_last_line = True
            found_already_deactivated_on_last_line = False
        else:
            if not found_already_deactivated_on_last_line:
                skipped_count += 1
            found_already_deactivated_on_last_line = True
            treated_last_line = False
        lines.append(line)
    text = ''.join(lines)
    if skipped is True:
        return text, count, skipped_count
    else:
        return text, count 



"""
Tools for modeling musical meter.
"""

import bisect
import collections
import uqbar.graphs
import abjad.rhythmtrees
from abjad import core
from abjad import indicators as abjad_indicators
from abjad import mathtools
from abjad import system
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.utilities.TypedCounter import TypedCounter
from abjad.utilities.TypedList import TypedList


class Meter(AbjadValueObject):
    """
    Meter.

    Meter models a common practice understanding of beats and other levels of
    rhythmic organization structured as a tree. Meter structure corresponds to
    the monotonically increasing sequence of factors in the numerator of a
    given time signature. Successively deeper levels of the tree divide time by
    successive factors.

    ..  container:: example

        Duple meter:

        >>> meter = abjad.Meter((2, 4))
        >>> meter
        Meter('(2/4 (1/4 1/4))')
        >>> print(meter.pretty_rtm_format)
        (2/4 (
            1/4
            1/4))

        >>> graph(meter) # doctest: +SKIP

        `2/4` comprises two beats.

    ..  container:: example

        Triple meter:

        >>> meter = abjad.Meter((3, 4))
        >>> print(meter.pretty_rtm_format)
        (3/4 (
            1/4
            1/4
            1/4))

        >>> graph(meter) # doctest: +SKIP

        `3/4` comprises three beats.

    ..  container:: example

        Quadruple meter:

        >>> meter = abjad.Meter((4, 4))
        >>> meter
        Meter('(4/4 (1/4 1/4 1/4 1/4))')
        >>> print(meter.pretty_rtm_format)
        (4/4 (
            1/4
            1/4
            1/4
            1/4))

        >>> graph(meter) # doctest: +SKIP

        `4/4` comprises four beats.

    ..  container:: example

        Compound triple meter:

        >>> meter = abjad.Meter((6, 8))
        >>> print(meter.pretty_rtm_format)
        (6/8 (
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))))

        >>> graph(meter) # doctest: +SKIP

        `6/8` comprises two beats of three parts each.

    ..  container:: example

        Another compound triple meter:

        >>> meter = abjad.Meter((12, 8))
        >>> print(meter.pretty_rtm_format)
        (12/8 (
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))))

        >>> graph(meter) # doctest: +SKIP

        `12/8` comprises four beats of three parts each.

    ..  container:: example

        An asymmetric meter:

        >>> meter = abjad.Meter((5, 4))
        >>> print(meter.pretty_rtm_format)
        (5/4 (
            (3/4 (
                1/4
                1/4
                1/4))
            (2/4 (
                1/4
                1/4))))

        >>> graph(meter) # doctest: +SKIP

        `5/4` comprises two unequal beats. By default unequal beats
        are arranged from greatest to least.

    ..  container:: example

        Another asymmetric meter:

        >>> meter = abjad.Meter((7, 4))
        >>> print(meter.pretty_rtm_format)
        (7/4 (
            (3/4 (
                1/4
                1/4
                1/4))
            (2/4 (
                1/4
                1/4))
            (2/4 (
                1/4
                1/4))))

        >>> graph(meter) # doctest: +SKIP

        `7/4` comprises three unequal beats. Beats are arranged from
        greatest to least by default.

    ..  container:: example

        The same asymmetric meter structured differently:

        >>> meter = abjad.Meter(
        ...     (7, 4),
        ...     decrease_monotonic=False,
        ...     )
        >>> print(meter.pretty_rtm_format)
        (7/4 (
            (2/4 (
                1/4
                1/4))
            (2/4 (
                1/4
                1/4))
            (3/4 (
                1/4
                1/4
                1/4))))

        >>> graph(meter) # doctest: +SKIP

        `7/4` with beats arragned from least to greatest.

    ..  container:: example

        Meter interpreted by default as containing two compound beats:

        >>> meter = abjad.Meter((6, 4))
        >>> meter
        Meter('(6/4 ((3/4 (1/4 1/4 1/4)) (3/4 (1/4 1/4 1/4))))')
        >>> print(meter.pretty_rtm_format)
        (6/4 (
            (3/4 (
                1/4
                1/4
                1/4))
            (3/4 (
                1/4
                1/4
                1/4))))

        >>> graph(meter) # doctest: +SKIP

        Same meter customized to contain four compound beats:

        >>> parser = abjad.rhythmtrees.RhythmTreeParser()
        >>> meter = abjad.Meter('(6/4 ((3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8))))')
        >>> meter
        Meter('(6/4 ((3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8)) (3/8 (1/8 1/8 1/8))))')
        >>> print(meter.pretty_rtm_format)
        (6/4 (
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))
            (3/8 (
                1/8
                1/8
                1/8))))

        >>> graph(meter) # doctest: +SKIP

    Prime divisions greater than ``3`` are converted to sequences of ``2``
    and ``3`` summing to that prime. Summands are arranged from greatest
    to least by default. This means that ``5`` becomes ``3+2`` and ``7``
    becomes ``3+2+2`` in the examples above.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_decrease_monotonic',
        '_denominator',
        '_numerator',
        '_preferred_boundary_depth',
        '_root_node',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        argument=None,
        decrease_monotonic=True,
        preferred_boundary_depth=None,
    ):

        argument = argument or (4, 4)
        assert isinstance(preferred_boundary_depth, (int, type(None)))
        self._preferred_boundary_depth = preferred_boundary_depth

        def recurse(
            node,
            factors,
            denominator,
            decrease_monotonic,
        ):
            if factors:
                factor, factors = factors[0], factors[1:]
                preprolated_duration = \
                    node.preprolated_duration.__div__(factor)
                #if factor in (2, 3, 4, 5):
                if factor in (2, 3, 4):
                    if factors:
                        for _ in range(factor):
                            child = abjad.rhythmtrees.RhythmTreeContainer(
                                preprolated_duration=preprolated_duration)
                            node.append(child)
                            recurse(
                                child,
                                factors,
                                denominator,
                                decrease_monotonic,
                                )
                    else:
                        for _ in range(factor):
                            node.append(
                                abjad.rhythmtrees.RhythmTreeLeaf(
                                    preprolated_duration=(1, denominator)))
                else:
                    parts = [3]
                    total = 3
                    while total < factor:
                        if decrease_monotonic:
                            parts.append(2)
                        else:
                            parts.insert(0, 2)
                        total += 2
                    for part in parts:
                        grouping = abjad.rhythmtrees.RhythmTreeContainer(
                            preprolated_duration=part * preprolated_duration)
                        if factors:
                            for _ in range(part):
                                child = abjad.rhythmtrees.RhythmTreeContainer(
                                    preprolated_duration=preprolated_duration)
                                grouping.append(child)
                                recurse(
                                    child,
                                    factors,
                                    denominator,
                                    decrease_monotonic,
                                    )
                        else:
                            for _ in range(part):
                                grouping.append(
                                    abjad.rhythmtrees.RhythmTreeLeaf(
                                        preprolated_duration=(1, denominator)))
                        node.append(grouping)
            else:
                node.extend([abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=(1, denominator))
                    for _ in range(node.preprolated_duration.numerator)])

        decrease_monotonic = bool(decrease_monotonic)

        try:
            numerator = argument.numerator
            denominator = argument.denominator
            is_fraction_like = True
        except AttributeError:
            is_fraction_like = False

        if isinstance(argument, type(self)):
            root = argument.root_node
            numerator, denominator = argument.numerator, argument.denominator
            decrease_monotonic = argument.decrease_monotonic

        elif isinstance(argument, (str, abjad.rhythmtrees.RhythmTreeContainer)):
            if isinstance(argument, str):
                parsed = abjad.rhythmtrees.RhythmTreeParser()(argument)
                assert len(parsed) == 1
                root = parsed[0]
            else:
                root = argument
            for node in [root] + list(root.depth_first()):
                assert node.prolation == 1
            numerator = root.preprolated_duration.numerator
            denominator = root.preprolated_duration.denominator

        elif (
            isinstance(argument, (tuple, core.Measure)) or
            is_fraction_like
        ):
            if isinstance(argument, tuple):
                fraction = mathtools.NonreducedFraction(argument)
            elif isinstance(argument, core.Measure):
                prototype = abjad_indicators.TimeSignature
                time_signature = argument._get_effective(prototype)
                fraction = mathtools.NonreducedFraction(
                    time_signature.numerator,
                    time_signature.denominator,
                    )
            else:
                fraction = mathtools.NonreducedFraction(
                    argument.numerator,
                    argument.denominator,
                    )
            numerator, denominator = fraction.numerator, fraction.denominator
            factors = mathtools.factors(numerator)
            # group two nested levels of 2s into a 4
            if 1 < len(factors) and factors[0] == factors[1] == 2:
                factors[0:2] = [4]
            root = abjad.rhythmtrees.RhythmTreeContainer(
                preprolated_duration=fraction)
            recurse(
                root,
                factors,
                denominator,
                decrease_monotonic,
                )

        else:
            message = 'can not initialize {}: {!r}.'
            message = message.format(type(self).__name__, argument)
            raise ValueError(message)

        self._root_node = root
        self._numerator = numerator
        self._denominator = denominator
        self._decrease_monotonic = decrease_monotonic

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a meter with an rtm format equal to that
        of this meter.

        Returns true or false.
        """
        return super().__eq__(argument)

    def __format__(self, format_specification=''):
        """
        Formats meter.

        ..  container:: example

            Gets storage format of ``7/4``:

            >>> meter = abjad.Meter((7, 4))
            >>> print(format(meter))
            abjad.Meter(
                '(7/4 ((3/4 (1/4 1/4 1/4)) (2/4 (1/4 1/4)) (2/4 (1/4 1/4))))'
                )

        Returns string.
        """
        if format_specification in ('', 'storage'):
            return system.StorageFormatManager(self).get_storage_format()
        return str(self)

    def __graph__(self, **keywords):
        """
        Gets Graphviz format of meter.

        ..  container:: example

            Graphs ``7/4``:

            >>> meter = abjad.Meter((7, 4))
            >>> meter_graph = meter.__graph__()
            >>> graph(meter_graph) # doctest: +SKIP

            ..  docs::

                >>> print(format(meter_graph, 'graphviz'))
                digraph G {
                    graph [bgcolor=transparent,
                        fontname=Arial,
                        penwidth=2,
                        truecolor=true];
                    node [fontname=Arial,
                        fontsize=12,
                        penwidth=2];
                    edge [penwidth=2];
                    node_0 [label="7/4",
                        shape=triangle];
                    node_1 [label="3/4",
                        shape=triangle];
                    node_2 [label="1/4",
                        shape=box];
                    node_3 [label="1/4",
                        shape=box];
                    node_4 [label="1/4",
                        shape=box];
                    node_5 [label="2/4",
                        shape=triangle];
                    node_6 [label="1/4",
                        shape=box];
                    node_7 [label="1/4",
                        shape=box];
                    node_8 [label="2/4",
                        shape=triangle];
                    node_9 [label="1/4",
                        shape=box];
                    node_10 [label="1/4",
                        shape=box];
                    subgraph cluster_offsets {
                        graph [style=rounded];
                        node_11_0 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 0 | <f_0_1> +++ }",
                            shape=Mrecord,
                            style=filled];
                        node_11_1 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 1/4 | <f_0_1> + }",
                            shape=Mrecord,
                            style=filled];
                        node_11_2 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 1/2 | <f_0_1> + }",
                            shape=Mrecord,
                            style=filled];
                        node_11_3 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 3/4 | <f_0_1> ++ }",
                            shape=Mrecord,
                            style=filled];
                        node_11_4 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 1 | <f_0_1> + }",
                            shape=Mrecord,
                            style=filled];
                        node_11_5 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 5/4 | <f_0_1> ++ }",
                            shape=Mrecord,
                            style=filled];
                        node_11_6 [color=white,
                            fillcolor=black,
                            fontcolor=white,
                            fontname="Arial bold",
                            label="{ <f_0_0> 3/2 | <f_0_1> + }",
                            shape=Mrecord,
                            style=filled];
                        node_11_7 [label="{ <f_0_0> 7/4 | <f_0_1> +++ }",
                            shape=Mrecord];
                    }
                    node_0 -> node_1;
                    node_0 -> node_5;
                    node_0 -> node_8;
                    node_1 -> node_2;
                    node_1 -> node_3;
                    node_1 -> node_4;
                    node_2 -> node_11_0 [style=dotted];
                    node_2 -> node_11_1 [style=dotted];
                    node_3 -> node_11_1 [style=dotted];
                    node_3 -> node_11_2 [style=dotted];
                    node_4 -> node_11_2 [style=dotted];
                    node_4 -> node_11_3 [style=dotted];
                    node_5 -> node_6;
                    node_5 -> node_7;
                    node_6 -> node_11_3 [style=dotted];
                    node_6 -> node_11_4 [style=dotted];
                    node_7 -> node_11_4 [style=dotted];
                    node_7 -> node_11_5 [style=dotted];
                    node_8 -> node_9;
                    node_8 -> node_10;
                    node_9 -> node_11_5 [style=dotted];
                    node_9 -> node_11_6 [style=dotted];
                    node_10 -> node_11_6 [style=dotted];
                    node_10 -> node_11_7 [style=dotted];
                }

        Returns Graphviz graph.
        """
        def make_offset_node(
            offset,
            leaf_one=None,
            leaf_two=None,
            is_last=False,
        ):
            if not is_last:
                offset_node = uqbar.graphs.Node(
                    attributes={
                        'shape': 'Mrecord',
                        'style': 'filled',
                        'color': 'white',
                        'fontname': 'Arial bold',
                        'fontcolor': 'white',
                        'fillcolor': 'black',
                        },
                    )
            else:
                offset_node = uqbar.graphs.Node(
                    attributes={
                        'shape': 'Mrecord',
                        },
                    )
            offset_field = uqbar.graphs.RecordField(
                label=str(offset),
                )
            weight_field = uqbar.graphs.RecordField(
                label='+' * offsets[offset],
                )
            group = uqbar.graphs.RecordGroup()
            group.extend([offset_field, weight_field])
            offset_node.append(group)
            offset_subgraph.append(offset_node)
            leaf_one_node = node_mapping[leaf_one]
            edge = uqbar.graphs.Edge(
                attributes={'style': 'dotted'},
                )
            edge.attach(leaf_one_node, offset_node)
            if leaf_two:
                leaf_two_node = node_mapping[leaf_two]
                edge = uqbar.graphs.Edge(
                    attributes={'style': 'dotted'},
                    )
                edge.attach(leaf_two_node, offset_node)
        import abjad
        offsets = abjad.MetricAccentKernel.count_offsets(
            abjad.sequence(self.depthwise_offset_inventory).flatten(depth=-1))
        graph = uqbar.graphs.Graph(
            name='G',
            attributes={
                'bgcolor': 'transparent',
                'fontname': 'Arial',
                'penwidth': 2,
                'truecolor': True,
                },
            edge_attributes={
                'penwidth': 2,
                },
            node_attributes={
                'fontname': 'Arial',
                'fontsize': 12,
                'penwidth': 2,
                },
            )
        node_mapping = {}
        root = self._root_node
        nodes = [root] + list(root.depth_first())
        leaves = [_ for _ in nodes if not hasattr(_, 'children')]
        for node in nodes:
            graphviz_node = uqbar.graphs.Node()
            graphviz_node.attributes['label'] = str(node.preprolated_duration)
            if isinstance(node, abjad.rhythmtrees.RhythmTreeContainer):
                graphviz_node.attributes['shape'] = 'triangle'
            else:
                graphviz_node.attributes['shape'] = 'box'
            graph.append(graphviz_node)
            node_mapping[node] = graphviz_node
            if node.parent is not None:
                uqbar.graphs.Edge().attach(
                    node_mapping[node.parent],
                    node_mapping[node],
                    )
        offset = leaves[0].start_offset
        offset_subgraph = uqbar.graphs.Graph(
            name='cluster_offsets',
            attributes={
                'style': 'rounded',
                },
            )
        graph.append(offset_subgraph)
        make_offset_node(offset, leaves[0])
        for one, two in abjad.sequence(leaves).nwise():
            offset = one.stop_offset
            make_offset_node(offset, one, two)
        offset = leaves[-1].stop_offset
        make_offset_node(offset, leaves[-1], is_last=True)
        return graph

    def __hash__(self):
        """
        Hashes meter.

        Returns integer.
        """
        return super().__hash__()

    def __iter__(self):
        """
        Iterates meter.

        ..  container:: example

            Iterates ``5/4``:


                >>> meter = abjad.Meter((5, 4))
                >>> for x in meter:
                ...    x
                ...
                (NonreducedFraction(0, 4), NonreducedFraction(1, 4))
                (NonreducedFraction(1, 4), NonreducedFraction(2, 4))
                (NonreducedFraction(2, 4), NonreducedFraction(3, 4))
                (NonreducedFraction(0, 4), NonreducedFraction(3, 4))
                (NonreducedFraction(3, 4), NonreducedFraction(4, 4))
                (NonreducedFraction(4, 4), NonreducedFraction(5, 4))
                (NonreducedFraction(3, 4), NonreducedFraction(5, 4))
                (NonreducedFraction(0, 4), NonreducedFraction(5, 4))

        Yields pairs.
        """
        def recurse(node):
            result = []
            for child in node:
                if isinstance(child, abjad.rhythmtrees.RhythmTreeLeaf):
                    result.append(child)
                else:
                    result.extend(recurse(child))
            result.append(node)
            return result
        result = recurse(self.root_node)
        for x in result:
            start_offset = mathtools.NonreducedFraction(
                x.start_offset).with_denominator(self.denominator)
            stop_offset = mathtools.NonreducedFraction(
                x.stop_offset).with_denominator(self.denominator)
            yield start_offset, stop_offset

    def __str__(self):
        """
        Gets string representation of meter.

        ..  container:: example

            Gets string representation of meters over ``8``:

            >>> for numerator in range(1, 9):
            ...     meter = abjad.Meter((numerator, 8))
            ...     print(str(meter))
            1/8
            2/8
            3/8
            4/8
            5/8
            6/8
            7/8
            8/8

        Returns string.
        """
        return '{}/{}'.format(self.numerator, self.denominator)

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return system.FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_args_values=[self.rtm_format],
            storage_format_kwargs_names=[],
            )

    @staticmethod
    def _make_gridded_test_rhythm(grid_length, rhythm_number, denominator=16):
        """
        Make test rhythm number ``rhythm_number`` that fits ``grid_length``.

        Returns selection of one or more possibly tied notes.

        ..  container:: example

            The eight test rhythms that fit a length-``4`` grid:

            >>> for rhythm_number in range(8):
            ...     notes = abjad.Meter._make_gridded_test_rhythm(
            ...         4, rhythm_number, denominator=4)
            ...     measure = abjad.Measure((4, 4), notes)
            ...     print('{}\t{}'.format(rhythm_number, str(measure)))
            ...
            0	Measure((4, 4), "c'1")
            1	Measure((4, 4), "c'2. c'4")
            2	Measure((4, 4), "c'2 c'4 c'4")
            3	Measure((4, 4), "c'2 c'2")
            4	Measure((4, 4), "c'4 c'4 c'2")
            5	Measure((4, 4), "c'4 c'4 c'4 c'4")
            6	Measure((4, 4), "c'4 c'2 c'4")
            7	Measure((4, 4), "c'4 c'2.")

        ..  container:: example

            The sixteenth test rhythms for that a length-``5`` grid:

            >>> for rhythm_number in range(16):
            ...     notes = abjad.Meter._make_gridded_test_rhythm(
            ...         5, rhythm_number, denominator=4)
            ...     measure = abjad.Measure((5, 4), notes)
            ...     print('{}\t{}'.format(rhythm_number, str(measure)))
            ...
            0	Measure((5, 4), "c'1 ~ c'4")
            1	Measure((5, 4), "c'1 c'4")
            2	Measure((5, 4), "c'2. c'4 c'4")
            3	Measure((5, 4), "c'2. c'2")
            4	Measure((5, 4), "c'2 c'4 c'2")
            5	Measure((5, 4), "c'2 c'4 c'4 c'4")
            6	Measure((5, 4), "c'2 c'2 c'4")
            7	Measure((5, 4), "c'2 c'2.")
            8	Measure((5, 4), "c'4 c'4 c'2.")
            9	Measure((5, 4), "c'4 c'4 c'2 c'4")
            10	Measure((5, 4), "c'4 c'4 c'4 c'4 c'4")
            11	Measure((5, 4), "c'4 c'4 c'4 c'2")
            12	Measure((5, 4), "c'4 c'2 c'2")
            13	Measure((5, 4), "c'4 c'2 c'4 c'4")
            14	Measure((5, 4), "c'4 c'2. c'4")
            15	Measure((5, 4), "c'4 c'1")

        Use for testing meter establishment.
        """
        import abjad
        # check input
        assert abjad.mathtools.is_positive_integer(grid_length)
        assert isinstance(rhythm_number, int)
        assert abjad.mathtools.is_positive_integer_power_of_two(denominator)
        # find count of all rhythms that fit grid length
        rhythm_count = 2 ** (grid_length - 1)
        # read rhythm number cyclically to allow large and
        # negative rhythm numbers
        rhythm_number = rhythm_number % rhythm_count
        # find binary representation of rhythm
        binary_representation = abjad.mathtools.integer_to_binary_string(
            rhythm_number)
        binary_representation = binary_representation.zfill(grid_length)
        # partition binary representation of rhythm
        parts = abjad.sequence(binary_representation).group_by()
        # find durations
        durations = [
            abjad.Duration(len(part), denominator)
            for part in parts
            ]
        # make notes
        maker = abjad.NoteMaker()
        notes = maker([0], durations)
        # return notes
        return notes

    @staticmethod
    def _rewrite_meter(
        components,
        meter,
        boundary_depth=None,
        initial_offset=None,
        maximum_dot_count=None,
        rewrite_tuplets=True,
        repeat_ties=False,
    ):
        def recurse(
            boundary_depth=None,
            boundary_offsets=None,
            depth=0,
            logical_tie=None,
        ):
            offsets = abjad.meter._MeterManager.get_offsets_at_depth(
                depth,
                offset_inventory,
                )
            #print('DEPTH:', depth)
            logical_tie_duration = logical_tie._get_preprolated_duration()
            logical_tie_timespan = abjad.inspect(logical_tie).timespan()
            logical_tie_start_offset = logical_tie_timespan.start_offset
            logical_tie_stop_offset = logical_tie_timespan.stop_offset
            logical_tie_starts_in_offsets = logical_tie_start_offset in offsets
            logical_tie_stops_in_offsets = logical_tie_stop_offset in offsets
            if not abjad.meter._MeterManager.is_acceptable_logical_tie(
                logical_tie_duration=logical_tie_duration,
                logical_tie_starts_in_offsets=logical_tie_starts_in_offsets,
                logical_tie_stops_in_offsets=logical_tie_stops_in_offsets,
                maximum_dot_count=maximum_dot_count,
            ):
                #print('UNACCEPTABLE:', logical_tie, logical_tie_start_offset, logical_tie_stop_offset)
                #print('\t', ' '.join([str(x) for x in offsets]))
                split_offset = None
                offsets = abjad.meter._MeterManager.get_offsets_at_depth(
                    depth,
                    offset_inventory,
                    )
                # If the logical tie's start aligns, take the latest possible offset.
                if logical_tie_starts_in_offsets:
                    offsets = reversed(offsets)
                for offset in offsets:
                    if logical_tie_start_offset < offset < logical_tie_stop_offset:
                        split_offset = offset
                        break
                #print('\tABS:', split_offset)
                if split_offset is not None:
                    split_offset -= logical_tie_start_offset
                    #print('\tREL:', split_offset)
                    #print()
                    shards = abjad.mutate(logical_tie[:]).split(
                        [split_offset],
                        repeat_ties=repeat_ties,
                        )
                    logical_ties = [abjad.LogicalTie(_) for _ in shards]
                    for logical_tie in logical_ties:
                        recurse(
                            boundary_depth=boundary_depth,
                            boundary_offsets=boundary_offsets,
                            depth=depth,
                            logical_tie=logical_tie,
                            )
                else:
                    #print()
                    recurse(
                        boundary_depth=boundary_depth,
                        boundary_offsets=boundary_offsets,
                        depth=depth + 1,
                        logical_tie=logical_tie,
                        )
            elif abjad.meter._MeterManager.is_boundary_crossing_logical_tie(
                boundary_depth=boundary_depth,
                boundary_offsets=boundary_offsets,
                logical_tie_start_offset=logical_tie_start_offset,
                logical_tie_stop_offset=logical_tie_stop_offset,
            ):
                #print('BOUNDARY CROSSING', logical_tie, logical_tie_start_offset, logical_tie_stop_offset)
                offsets = boundary_offsets
                if logical_tie_start_offset in boundary_offsets:
                    offsets = reversed(boundary_offsets)
                split_offset = None
                for offset in offsets:
                    if logical_tie_start_offset < offset < logical_tie_stop_offset:
                        split_offset = offset
                        break
                assert split_offset is not None
                #print('\tABS:', split_offset)
                split_offset -= logical_tie_start_offset
                #print('\tREL:', split_offset)
                #print()
                shards = abjad.mutate(logical_tie[:]).split(
                    [split_offset],
                    repeat_ties=repeat_ties,
                    )
                logical_ties = [abjad.LogicalTie(shard) for shard in shards]
                for logical_tie in logical_ties:
                    recurse(
                        boundary_depth=boundary_depth,
                        boundary_offsets=boundary_offsets,
                        depth=depth,
                        logical_tie=logical_tie,
                        )
            else:
                #print('ACCEPTABLE:', logical_tie, logical_tie_start_offset, logical_tie_stop_offset)
                #print('\t', ' '.join([str(x) for x in offsets]))
                #print()
                logical_tie[:]._fuse()
        import abjad
        assert isinstance(components, abjad.Selection), repr(components)
        if not isinstance(meter, abjad.Meter):
            meter = abjad.Meter(meter)
        boundary_depth = boundary_depth or meter.preferred_boundary_depth
        # Validate arguments.
        assert abjad.select(components).are_contiguous_logical_voice()
        if not isinstance(meter, abjad.Meter):
            meter = abjad.Meter(meter)
        if boundary_depth is not None:
            boundary_depth = int(boundary_depth)
        if maximum_dot_count is not None:
            maximum_dot_count = int(maximum_dot_count)
            assert 0 <= maximum_dot_count
        if initial_offset is None:
            initial_offset = abjad.Offset(0)
        initial_offset = abjad.Offset(initial_offset)
        first_start_offset = abjad.inspect(
            components[0]).timespan().start_offset
        last_start_offset = abjad.inspect(
            components[-1]).timespan().start_offset
        difference = last_start_offset - first_start_offset + initial_offset
        assert difference < meter.implied_time_signature.duration
        # Build offset inventory, adjusted for initial offset and prolation.
        first_offset = abjad.inspect(components[0]).timespan().start_offset
        first_offset -= initial_offset
        prolation = abjad.inspect(components[0]).parentage(
            include_self=False).prolation
        offset_inventory = []
        for offsets in meter.depthwise_offset_inventory:
            offsets = [(x * prolation) + first_offset for x in offsets]
            offset_inventory.append(tuple(offsets))
        # Build boundary offset inventory, if applicable.
        if boundary_depth is not None:
            boundary_offsets = offset_inventory[boundary_depth]
        else:
            boundary_offsets = None
        # Cache results of iterator, as we'll be mutating the underlying collection
        iterator = abjad.meter._MeterManager.iterate_rewrite_inputs(components)
        items = tuple(iterator)
        for item in items:
            if isinstance(item, abjad.LogicalTie):
                #print('RECURSING:', item)
                recurse(
                    boundary_depth=boundary_depth,
                    boundary_offsets=boundary_offsets,
                    depth=0,
                    logical_tie=item,
                    )
            elif isinstance(item, abjad.Tuplet) and not rewrite_tuplets:
                pass
            else:
                #print('DESCENDING:', item)
                preprolated_duration = sum(
                    [x._get_preprolated_duration() for x in item]
                    )
                if preprolated_duration.numerator == 1:
                    preprolated_duration = abjad.NonreducedFraction(
                        preprolated_duration)
                    preprolated_duration = preprolated_duration.with_denominator(
                        preprolated_duration.denominator * 4)
                sub_metrical_hierarchy = abjad.Meter(preprolated_duration)
                sub_boundary_depth = 1
                if boundary_depth is None:
                    sub_boundary_depth = None
                Meter._rewrite_meter(
                    item[:],
                    sub_metrical_hierarchy,
                    boundary_depth=sub_boundary_depth,
                    maximum_dot_count=maximum_dot_count,
                    )

    ### PUBLIC METHODS ###

    @staticmethod
    def fit_meters(
        argument,
        meters,
        denominator=32,
        discard_final_orphan_downbeat=True,
        maximum_run_length=None,
        starting_offset=None,
    ):
        """
        Finds the best-matching sequence of meters for the offsets
        contained in ``argument``.

        ..  container:: example

            >>> meters = [(3, 4), (4, 4), (5, 4)]
            >>> meters = [abjad.Meter(_) for _ in meters]

        ..  container:: example

            Matches a series of hypothetical ``4/4`` measures:

            >>> argument = [(0, 4), (4, 4), (8, 4), (12, 4), (16, 4)]
            >>> for x in abjad.Meter.fit_meters(
            ...     argument, meters):
            ...     print(x.implied_time_signature)
            ...
            4/4
            4/4
            4/4
            4/4

        ..  container:: example

            Matches a series of hypothetical ``5/4`` measures:

            >>> argument = [(0, 4), (3, 4), (5, 4), (10, 4), (15, 4), (20, 4)]
            >>> for x in abjad.Meter.fit_meters(
            ...     argument, meters):
            ...     print(x.implied_time_signature)
            ...
            3/4
            4/4
            3/4
            5/4
            5/4

        Coerces offsets from ``argument`` via
        ``MetricAccentKernel.count_offsets()``.

        Coerces Meters from ``meters`` via ``MeterList``.

        Returns list.
        """
        import abjad
        session = abjad.meter._MeterFittingSession(
            kernel_denominator=denominator,
            maximum_run_length=maximum_run_length,
            meters=meters,
            offset_counter=argument,
            )
        meters = session()
        return meters

    def generate_offset_kernel_to_denominator(
        self,
        denominator,
        normalize=True,
    ):
        r"""
        Generates a dictionary of all offsets in a meter up
        to ``denominator``.

        Keys are the offsets and the values are the normalized weights of
        those offsets.

        ..  container:: example

            >>> meter = abjad.Meter((4, 4))
            >>> kernel = meter.generate_offset_kernel_to_denominator(8)
            >>> for offset, weight in sorted(kernel.kernel.items()):
            ...     print('{!s}\t{!s}'.format(offset, weight))
            ...
            0       3/16
            1/8     1/16
            1/4     1/8
            3/8     1/16
            1/2     1/8
            5/8     1/16
            3/4     1/8
            7/8     1/16
            1       3/16

        This is useful for testing how strongly a collection of offsets
        responds to a given meter.

        Returns dictionary.
        """
        import abjad
        assert mathtools.is_positive_integer_power_of_two(
            denominator // self.denominator)
        inventory = list(self.depthwise_offset_inventory)
        old_flag_count = abjad.Duration(1, self.denominator).flag_count
        new_flag_count = abjad.Duration(1, denominator).flag_count
        extra_depth = new_flag_count - old_flag_count
        for _ in range(extra_depth):
            old_offsets = inventory[-1]
            new_offsets = []
            for first, second in abjad.sequence(old_offsets).nwise():
                new_offsets.append(first)
                new_offsets.append((first + second) / 2)
            new_offsets.append(old_offsets[-1])
            inventory.append(tuple(new_offsets))
        total = 0
        kernel = {}
        for offsets in inventory:
            for offset in offsets:
                if offset not in kernel:
                    kernel[offset] = 0
                kernel[offset] += 1
                total += 1
        if normalize:
            for offset, response in kernel.items():
                kernel[offset] = abjad.Multiplier(response, total)
        return abjad.MetricAccentKernel(kernel)

    ### PUBLIC PROPERTIES ###

    @property
    def decrease_monotonic(self):
        """
        Is true when meter divides large primes into collections of ``2``
        and ``3`` that decrease monotonically.

        ..  container:: example

            An asymmetric meter with beats arranged greatest to least:

            >>> meter = abjad.Meter(
            ...     (7, 4),
            ...     decrease_monotonic=True,
            ...     )

            >>> meter.decrease_monotonic
            True

            >>> print(meter.pretty_rtm_format)
            (7/4 (
                (3/4 (
                    1/4
                    1/4
                    1/4))
                (2/4 (
                    1/4
                    1/4))
                (2/4 (
                    1/4
                    1/4))))

            This is default beahvior.

        ..  container:: example

            The same asymmetric meter with unequal beats arranged least to
            greatest:

            >>> meter = abjad.Meter(
            ...     (7, 4),
            ...     decrease_monotonic=False,
            ...     )

            >>> meter.decrease_monotonic
            False

            >>> print(meter.pretty_rtm_format)
            (7/4 (
                (2/4 (
                    1/4
                    1/4))
                (2/4 (
                    1/4
                    1/4))
                (3/4 (
                    1/4
                    1/4
                    1/4))))

        Returns true or false.
        """
        return self._decrease_monotonic

    @property
    def denominator(self):
        """
        Gets denominator of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> meter.denominator
            4

        Returns positive integer.
        """
        return self._denominator

    @property
    def depthwise_offset_inventory(self):
        """
        Gets depthwise offset inventory of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> for depth, offsets in enumerate(
            ...     meter.depthwise_offset_inventory):
            ...     print(depth, offsets)
            0 (Offset(0, 1), Offset(7, 4))
            1 (Offset(0, 1), Offset(3, 4), Offset(5, 4), Offset(7, 4))
            2 (Offset(0, 1), Offset(1, 4), Offset(1, 2), Offset(3, 4), Offset(1, 1), Offset(5, 4), Offset(3, 2), Offset(7, 4))

        Returns dictionary.
        """
        import abjad
        inventory = []
        all_offsets = set()
        all_offsets.add(abjad.Offset(self.numerator, self.denominator))
        for depth, nodes in sorted(self.root_node._depthwise_inventory.items()):
            for node in nodes:
                all_offsets.add(abjad.Offset(node.start_offset))
            inventory.append(tuple(sorted(all_offsets)))
        return tuple(inventory)

    @property
    def duration(self):
        """
        Gets duration of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> meter.duration
            Duration(7, 4)

        Returns duration.
        """
        import abjad
        return abjad.Duration(self.numerator, self.denominator)

    @property
    def implied_time_signature(self):
        """
        Gets implied time signature of meter.

        ..  container:: example

            >>> abjad.Meter((4, 4)).implied_time_signature
            TimeSignature((4, 4))

        Returns time signature.
        """
        return abjad_indicators.TimeSignature(
            self.root_node.preprolated_duration
            )

    @property
    def is_compound(self):
        """
        Is true when meter is compound.

        ..  container:: example

            Compound meters written over ``4``:

            >>> for numerator in range(1, 13):
            ...     meter = abjad.Meter((numerator, 4))
            ...     string = True if meter.is_compound else ''
            ...     print(str(meter), string)
            ...
            1/4
            2/4
            3/4
            4/4
            5/4
            6/4     True
            7/4
            8/4
            9/4     True
            10/4
            11/4
            12/4    True

        ..  container:: example

            Compound meters written over ``8``:

            >>> for numerator in range(1, 13):
            ...     meter = abjad.Meter((numerator, 8))
            ...     string = True if meter.is_compound else ''
            ...     print(str(meter), string)
            ...
            1/8
            2/8
            3/8
            4/8
            5/8
            6/8     True
            7/8
            8/8
            9/8     True
            10/8
            11/8
            12/8    True

        Compound meters defined equal to those meters with a numerator
        divisible by ``3`` (but not equal to ``3``).

        Returns true or false.
        """
        if 3 in mathtools.divisors(self.numerator):
            if not self.numerator == 3:
                return True
        return False

    @property
    def is_simple(self):
        """
        Is true when meter is simple.

        ..  container:: example

            Simple meters written over ``4``:

            >>> for numerator in range(1, 13):
            ...     meter = abjad.Meter((numerator, 4))
            ...     string = True if meter.is_simple else ''
            ...     print(str(meter), string)
            ...
            1/4     True
            2/4     True
            3/4     True
            4/4     True
            5/4     True
            6/4
            7/4     True
            8/4     True
            9/4
            10/4    True
            11/4    True
            12/4

        ..  container:: example

            Simple meters written over ``8``:

            >>> for numerator in range(1, 13):
            ...     meter = abjad.Meter((numerator, 8))
            ...     string = True if meter.is_simple else ''
            ...     print(str(meter), string)
            ...
            1/8     True
            2/8     True
            3/8     True
            4/8     True
            5/8     True
            6/8
            7/8     True
            8/8     True
            9/8
            10/8    True
            11/8    True
            12/8

        Simple meters defined equal to those meters with a numerator
        not divisible by ``3``.

        Meters with numerator equal to ``3`` are also defined as simple.

        Returns true or false.
        """
        return not self.is_compound

    @property
    def numerator(self):
        """
        Gets numerator of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> meter.numerator
            7

        Returns positive integer.
        """
        return self._numerator

    @property
    def pair(self):
        """
        Gets pair of numerator and denominator of meter.

        ..  container:: example

            >>> meter = abjad.Meter((6, 4))
            >>> meter.pair
            (6, 4)

        Returns pair.
        """
        return (self.numerator, self.denominator)

    @property
    def preferred_boundary_depth(self):
        """
        Gets preferred boundary depth of meter.

        ..  container:: example

            No preferred boundary depth:

            >>> abjad.Meter((6, 8)).preferred_boundary_depth is None
            True

        ..  container:: example

            Customized preferred boundary depth:

            >>> meter = abjad.Meter(
            ...     (6, 8),
            ...     preferred_boundary_depth=1,
            ...     )
            >>> meter.preferred_boundary_depth
            1

        Used by ``mutate().rewrite_meter()``.

        Defaults to none.

        Set to integer or none.

        Returns integer or none.
        """
        return self._preferred_boundary_depth

    @property
    def pretty_rtm_format(self):
        """
        Gets pretty RTM format of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> print(meter.pretty_rtm_format)
            (7/4 (
                (3/4 (
                    1/4
                    1/4
                    1/4))
                (2/4 (
                    1/4
                    1/4))
                (2/4 (
                    1/4
                    1/4))))

        Returns string.
        """
        return self.root_node.pretty_rtm_format

    @property
    def root_node(self):
        """
        Gets root node of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> print(format(meter.root_node))
            abjad.rhythmtrees.RhythmTreeContainer(
                children=(
                    abjad.rhythmtrees.RhythmTreeContainer(
                        children=(
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            ),
                        preprolated_duration=abjad.NonreducedFraction(3, 4),
                        ),
                    abjad.rhythmtrees.RhythmTreeContainer(
                        children=(
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            ),
                        preprolated_duration=abjad.NonreducedFraction(2, 4),
                        ),
                    abjad.rhythmtrees.RhythmTreeContainer(
                        children=(
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            abjad.rhythmtrees.RhythmTreeLeaf(
                                preprolated_duration=abjad.Duration(1, 4),
                                is_pitched=True,
                                ),
                            ),
                        preprolated_duration=abjad.NonreducedFraction(2, 4),
                        ),
                    ),
                preprolated_duration=abjad.NonreducedFraction(7, 4),
                )

        Returns rhythm tree node.
        """
        return self._root_node

    @property
    def rtm_format(self):
        """
        Gets RTM format of meter.

        ..  container:: example

            >>> meter = abjad.Meter((7, 4))
            >>> meter.rtm_format
            '(7/4 ((3/4 (1/4 1/4 1/4)) (2/4 (1/4 1/4)) (2/4 (1/4 1/4))))'

        Returns string.
        """
        return self._root_node.rtm_format


class MeterList(TypedList):
    """
    Meter list.

    ..  container:: example

        >>> meters = abjad.MeterList([
        ...     (3, 4), (5, 16), (7, 8),
        ...     ])

        >>> abjad.f(meters)
        abjad.MeterList(
            [
                abjad.Meter(
                    '(3/4 (1/4 1/4 1/4))'
                    ),
                abjad.Meter(
                    '(5/16 ((3/16 (1/16 1/16 1/16)) (2/16 (1/16 1/16))))'
                    ),
                abjad.Meter(
                    '(7/8 ((3/8 (1/8 1/8 1/8)) (2/8 (1/8 1/8)) (2/8 (1/8 1/8))))'
                    ),
                ]
            )

        >>> abjad.show(meters, scale=0.5) # doctest: +SKIP

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __illustrate__(self, denominator=16, range_=None, scale=None):
        r"""
        Illustrates meters.

        ..  container:: example

            >>> meters = abjad.MeterList([
            ...     (3, 4), (5, 16), (7, 8),
            ...     ])
            >>> abjad.show(meters, scale=0.5) # doctest: +SKIP

            ..  doctest

                >>> lilypond_file = meters.__illustrate__()
                >>> abjad.f(lilypond_file) # doctest: +SKIP
                \version "2.19..."
                \language "english"
                <BLANKLINE>
                \header {
                    tagline = ##f
                }
                <BLANKLINE>
                \layout {}
                <BLANKLINE>
                \paper {}
                <BLANKLINE>
                \markup {
                    \column
                        {
                            \combine
                                \combine
                                    \translate
                                        #'(1.0 . 1)
                                        \sans
                                            \fontsize
                                                #-3
                                                \center-align
                                                    \fraction
                                                        3
                                                        4
                                    \translate
                                        #'(49.38709677419355 . 1)
                                        \sans
                                            \fontsize
                                                #-3
                                                \center-align
                                                    \fraction
                                                        5
                                                        16
                                \translate
                                    #'(69.54838709677419 . 1)
                                    \sans
                                        \fontsize
                                            #-3
                                            \center-align
                                                \fraction
                                                    7
                                                    8
                            \combine
                                \postscript
                                    #"
                                    0.2 setlinewidth
                                    1 0.5 moveto
                                    49.387... 0.5 lineto
                                    stroke
                                    1 1.25 moveto
                                    1 -0.25 lineto
                                    stroke
                                    49.387... 1.25 moveto
                                    49.387... -0.25 lineto
                                    stroke
                                    49.387... 0.5 moveto
                                    69.548... 0.5 lineto
                                    stroke
                                    49.387... 1.25 moveto
                                    49.387... -0.25 lineto
                                    stroke
                                    69.548... 1.25 moveto
                                    69.548... -0.25 lineto
                                    stroke
                                    69.5483... 0.5 moveto
                                    126 0.5 lineto
                                    stroke
                                    69.5483... 1.25 moveto
                                    69.5483... -0.25 lineto
                                    stroke
                                    126 1.25 moveto
                                    126 -0.25 lineto
                                    stroke
                                    "
                                \postscript
                                    #"
                                    1 -2 moveto
                                    0 -6.153... rlineto
                                    stroke
                                    5.0322... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    9.0645... -2 moveto
                                    0 -3.076... rlineto
                                    stroke
                                    13.0967... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    17.129... -2 moveto
                                    0 -4.615... rlineto
                                    stroke
                                    21.161... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    25.193... -2 moveto
                                    0 -3.076... rlineto
                                    stroke
                                    29.225... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    33.258... -2 moveto
                                    0 -4.615... rlineto
                                    stroke
                                    37.290... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    41.322... -2 moveto
                                    0 -3.076... rlineto
                                    stroke
                                    45.354... -2 moveto
                                    0 -1.538... rlineto
                                    stroke
                                    49.387... -2 moveto
                                    0 -6.153... rlineto
                                    stroke
                                    49.387... -2 moveto
                                    0 -10.909... rlineto
                                    stroke
                                    53.419... -2 moveto
                                    0 -3.636... rlineto
                                    stroke
                                    57.451... -2 moveto
                                    0 -3.636... rlineto
                                    stroke
                                    61.483... -2 moveto
                                    0 -7.272... rlineto
                                    stroke
                                    65.516... -2 moveto
                                    0 -3.636... rlineto
                                    stroke
                                    69.548... -2 moveto
                                    0 -10.909... rlineto
                                    stroke
                                    69.548... -2 moveto
                                    0 -5.517... rlineto
                                    stroke
                                    73.580... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    77.612... -2 moveto
                                    0 -2.758... rlineto
                                    stroke
                                    81.645... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    85.677... -2 moveto
                                    0 -2.758... rlineto
                                    stroke
                                    89.709... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    93.741... -2 moveto
                                    0 -4.137... rlineto
                                    stroke
                                    97.774... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    101.806... -2 moveto
                                    0 -2.758... rlineto
                                    stroke
                                    105.838... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    109.870... -2 moveto
                                    0 -4.137... rlineto
                                    stroke
                                    113.903... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    117.935... -2 moveto
                                    0 -2.758... rlineto
                                    stroke
                                    121.967... -2 moveto
                                    0 -1.379... rlineto
                                    stroke
                                    126 -2 moveto
                                    0 -5.517... rlineto
                                    stroke
                                    "
                        }
                    }

        Returns LilyPond file.
        """
        import abjad
        durations = [_.duration for _ in self]
        total_duration = sum(durations)
        offsets = abjad.mathtools.cumulative_sums(durations, start=0)
        timespans = abjad.TimespanList()
        for one, two in abjad.sequence(offsets).nwise():
            timespan = abjad.Timespan(
                start_offset=one,
                stop_offset=two,
                )
            timespans.append(timespan)
        if range_ is not None:
            minimum, maximum = range_
        else:
            minimum, maximum = 0, total_duration
        minimum = float(abjad.Offset(minimum))
        maximum = float(abjad.Offset(maximum))
        if scale is None:
            scale = 1.
        assert 0 < scale
        postscript_scale = 125. / (maximum - minimum)
        postscript_scale *= float(scale)
        postscript_x_offset = (minimum * postscript_scale) - 1
        timespan_markup = timespans._make_timespan_list_markup(
            timespans,
            postscript_x_offset,
            postscript_scale,
            draw_offsets=False,
            )
        ps = abjad.Postscript()
        rational_x_offset = abjad.Offset(0)
        for meter in self:
            kernel_denominator = denominator or meter.denominator
            kernel = abjad.meter.MetricAccentKernel.from_meter(
                meter, kernel_denominator)
            for offset, weight in sorted(kernel.kernel.items()):
                weight = float(weight) * -40
                ps_x_offset = float(rational_x_offset + offset)
                ps_x_offset *= postscript_scale
                ps_x_offset += 1
                ps = ps.moveto(ps_x_offset, -2)
                ps = ps.rlineto(0, weight)
                ps = ps.stroke()
            rational_x_offset += meter.duration
        ps = abjad.Markup.postscript(ps)
        markup_list = abjad.MarkupList([timespan_markup, ps])
        lines_markup = markup_list.combine()
        fraction_markups = []
        for meter, offset in zip(self, offsets):
            numerator, denominator = meter.numerator, meter.denominator
            fraction = abjad.Markup.fraction(numerator, denominator)
            fraction = fraction.center_align().fontsize(-3).sans()
            x_translation = (float(offset) * postscript_scale)
            x_translation -= postscript_x_offset
            fraction = fraction.translate((x_translation, 1))
            fraction_markups.append(fraction)
        fraction_markup = fraction_markups[0]
        for markup in fraction_markups[1:]:
            markup_list = [fraction_markup, markup]
            markup_list = abjad.MarkupList(markup_list)
            fraction_markup = markup_list.combine()
        markup = abjad.Markup.column([fraction_markup, lines_markup])
        return markup.__illustrate__()

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        import abjad
        return abjad.meter.Meter


class MetricAccentKernel(AbjadValueObject):
    """
    Metric accent kernel.

    ..  container:: example

        >>> hierarchy = abjad.Meter((7, 8))
        >>> kernel = hierarchy.generate_offset_kernel_to_denominator(8)
        >>> kernel
        MetricAccentKernel(
            {
                Offset(0, 1): Multiplier(3, 14),
                Offset(1, 8): Multiplier(1, 14),
                Offset(1, 4): Multiplier(1, 14),
                Offset(3, 8): Multiplier(1, 7),
                Offset(1, 2): Multiplier(1, 14),
                Offset(5, 8): Multiplier(1, 7),
                Offset(3, 4): Multiplier(1, 14),
                Offset(7, 8): Multiplier(3, 14),
                }
            )

    Call the kernel against an expression from which offsets can be counted
    to receive an impulse-response:

    ..  container:: example

        >>> offsets = [(0, 8), (1, 8), (1, 8), (3, 8)]
        >>> kernel(offsets)
        Multiplier(1, 2)

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_kernel',
        '_offsets',
        )

    ### INITIALIZER ###

    def __init__(self, kernel=None):
        import abjad
        kernel = kernel or {}
        assert isinstance(kernel, dict)
        #assert 1 < len(kernel)
        for key, value in kernel.items():
            assert isinstance(key, abjad.Offset)
            assert isinstance(value, abjad.Multiplier)
        self._kernel = kernel.copy()
        self._offsets = tuple(sorted(self._kernel))

    ### SPECIAL METHODS ###

    def __call__(self, argument):
        """
        Calls metrical accent kernal on ``argument``.

        >>> upper_staff = abjad.Staff("c'8 d'4. e'8 f'4.")
        >>> lower_staff = abjad.Staff(r'\clef bass c4 b,4 a,2')
        >>> score = abjad.Score([upper_staff, lower_staff])

        >>> kernel = abjad.MetricAccentKernel.from_meter((4, 4))
        >>> kernel(score)
        Multiplier(10, 33)

        Returns float.
        """
        import abjad
        offset_count = self.count_offsets(argument)
        response = abjad.Multiplier(0, 1)
        for offset, count in offset_count.items():
            if offset in self._kernel:
                weight = self._kernel[offset]
                weighted_count = weight * count
                response += weighted_count
        return response

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a metrical accent kernal with a kernal
        equal to that of this metrical accent kernel.

        Returns true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes metric accent kernel.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return system.FormatSpecification(
            client=self,
            repr_is_indented=True,
            storage_format_args_values=[self.kernel],
            storage_format_kwargs_names=[],
            )

    ### PUBLIC METHODS ###

    @staticmethod
    def count_offsets(argument):
        r"""
        Count offsets in ``argument``.

        ..  container:: example

            >>> upper_staff = abjad.Staff("c'8 d'4. e'8 f'4.")
            >>> lower_staff = abjad.Staff(r'\clef bass c4 b,4 a,2')
            >>> score = abjad.Score([upper_staff, lower_staff])

            ..  docs::

                >>> abjad.f(score)
                \new Score
                <<
                    \new Staff
                    {
                        c'8
                        d'4.
                        e'8
                        f'4.
                    }
                    \new Staff
                    {
                        \clef "bass"
                        c4
                        b,4
                        a,2
                    }
                >>

            >>> abjad.show(score) # doctest: +SKIP

            >>> MetricAccentKernel = abjad.MetricAccentKernel
            >>> leaves = abjad.select(score).leaves()
            >>> counter = abjad.MetricAccentKernel.count_offsets(leaves)
            >>> for offset, count in sorted(counter.items()):
            ...     offset, count
            ...
            (Offset(0, 1), 2)
            (Offset(1, 8), 2)
            (Offset(1, 4), 2)
            (Offset(1, 2), 4)
            (Offset(5, 8), 2)
            (Offset(1, 1), 2)

        ..  container:: example

            >>> a = abjad.Timespan(0, 10)
            >>> b = abjad.Timespan(5, 15)
            >>> c = abjad.Timespan(15, 20)

            >>> counter = MetricAccentKernel.count_offsets((a, b, c))
            >>> for offset, count in sorted(counter.items()):
            ...     offset, count
            ...
            (Offset(0, 1), 1)
            (Offset(5, 1), 1)
            (Offset(10, 1), 1)
            (Offset(15, 1), 2)
            (Offset(20, 1), 1)

        Returns counter.
        """
        from abjad import meter
        return meter.OffsetCounter(argument)

    @staticmethod
    def from_meter(meter, denominator=32, normalize=True):
        """
        Create a metric accent kernel from ``meter``.

        Returns new metric accent kernel.
        """
        import abjad
        if not isinstance(meter, abjad.Meter):
            meter = abjad.Meter(meter)
        return meter.generate_offset_kernel_to_denominator(
            denominator=denominator,
            normalize=normalize,
            )

    ### PUBLIC PROPERTIES ###

    @property
    def duration(self):
        """
        Gets duration.
        """
        import abjad
        return abjad.Duration(self._offsets[-1])

    @property
    def kernel(self):
        """
        The kernel datastructure.

        Returns dict.
        """
        return self._kernel.copy()


class OffsetCounter(TypedCounter):
    """
    Offset counter.

    ..  container:: example

        >>> timespans = abjad.TimespanList([
        ...     abjad.Timespan(0, 16),
        ...     abjad.Timespan(5, 12),
        ...     abjad.Timespan(-2, 8),
        ...     ])
        >>> timespan_operand = abjad.Timespan(6, 10)
        >>> timespans = timespans - timespan_operand
        >>> offset_counter = abjad.OffsetCounter(timespans)

        >>> abjad.f(offset_counter)
        abjad.OffsetCounter(
            {
                abjad.Offset(-2, 1): 1,
                abjad.Offset(0, 1): 1,
                abjad.Offset(5, 1): 1,
                abjad.Offset(6, 1): 3,
                abjad.Offset(10, 1): 2,
                abjad.Offset(12, 1): 1,
                abjad.Offset(16, 1): 1,
                }
            )

        >>> abjad.show(offset_counter, scale=0.5) # doctest: +SKIP

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None):
        import abjad
        TypedCounter.__init__(self, item_class=abjad.Offset)
        if items:
            for item in items:
                try:
                    self[item.start_offset] += 1
                    self[item.stop_offset] += 1
                except Exception:
                    if hasattr(item, '_get_timespan'):
                        self[abjad.inspect(item).timespan().start_offset] += 1
                        self[abjad.inspect(item).timespan().stop_offset] += 1
                    else:
                        offset = abjad.Offset(item)
                        self[offset] += 1

    ### SPECIAL METHODS ###

    def __illustrate__(self, range_=None, scale=None):
        r"""
        Illustrates offset counter.

        ..  container:: example

            >>> timespans = abjad.TimespanList([
            ...     abjad.Timespan(0, 16),
            ...     abjad.Timespan(5, 12),
            ...     abjad.Timespan(-2, 8),
            ...     ])
            >>> timespan_operand = abjad.Timespan(6, 10)
            >>> timespans = timespans - timespan_operand
            >>> offset_counter = abjad.OffsetCounter(timespans)
            >>> abjad.show(offset_counter, scale=0.5) # doctest: +SKIP

        Returns LilyPond file.
        """
        import abjad
        if not self:
            return abjad.Markup.null().__illustrate__()
        if isinstance(range_, abjad.Timespan):
            minimum, maximum = range_.start_offset, range_.stop_offset
        elif range_ is not None:
            minimum, maximum = range_
        else:
            minimum, maximum = min(self), max(self)
        minimum = float(abjad.Offset(minimum))
        maximum = float(abjad.Offset(maximum))
        if scale is None:
            scale = 1.
        assert 0 < scale
        postscript_scale = 150. / (maximum - minimum)
        postscript_scale *= float(scale)
        postscript_x_offset = (minimum * postscript_scale) - 1
        ps = abjad.Postscript()
        ps = ps.setlinewidth(0.2)
        ps = ps.setdash([2, 1])
        for offset, count in sorted(self.items()):
            offset = (float(offset) * postscript_scale)
            offset -= postscript_x_offset
            ps = ps.moveto(offset, -1)
            ps = ps.rlineto(0, (float(count) * -3) + 1)
            ps = ps.stroke()
        markup = abjad.Markup.postscript(ps)
        pieces = [markup]
        for offset in sorted(self):
            offset = abjad.Multiplier(offset)
            numerator, denominator = offset.numerator, offset.denominator
            fraction = abjad.Markup.fraction(numerator, denominator)
            fraction = fraction.center_align().fontsize(-3).sans()
            x_translation = (float(offset) * postscript_scale)
            x_translation -= postscript_x_offset
            fraction = fraction.translate((x_translation, 1))
            pieces.append(fraction)
        markup = abjad.Markup.overlay(pieces)
        return markup.__illustrate__()

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        import abjad
        return abjad.Offset


class _MeterFittingSession(AbjadValueObject):
    """
    Meter-fitting session.

    Used internally by Meter.fit_meters().
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_cached_offset_counters',
        '_kernel_denominator',
        '_kernels',
        '_longest_kernel',
        '_maximum_run_length',
        '_meters',
        '_offset_counter',
        '_ordered_offsets',
        )

    KernelScore = collections.namedtuple(
        'KernelScore',
        ('kernel', 'score'),
        )

    ### INITIALIZER ###

    def __init__(
        self,
        kernel_denominator=32,
        maximum_run_length=None,
        meters=None,
        offset_counter=None,
    ):
        import abjad
        self._cached_offset_counters = {}
        if maximum_run_length is not None:
            maximum_run_length = int(maximum_run_length)
            assert 0 < maximum_run_length
        self._maximum_run_length = maximum_run_length
        if offset_counter:
            self._offset_counter = abjad.MetricAccentKernel.count_offsets(
                offset_counter)
        else:
            self._offset_counter = {}
        self._ordered_offsets = tuple(sorted(self.offset_counter))
        meters = meters or ()
        self._meters = tuple(abjad.Meter(_) for _ in meters)
        self._kernel_denominator = abjad.Duration(kernel_denominator)
        self._kernels = {}
        for meter in self._meters:
            kernel = meter.generate_offset_kernel_to_denominator(
                self._kernel_denominator)
            self._kernels[kernel] = meter
        if self.kernels:
            self._longest_kernel = sorted(
                self._kernels,
                key=lambda x: x.duration,
                )[-1]
        else:
            self._longest_kernel = None

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Fits meters.

        Returns meter list.
        """
        import abjad
        selected_kernels = []
        current_offset = abjad.Offset(0)
        while current_offset < self.ordered_offsets[-1]:
            kernel_scores = []
            kernels = self._get_kernels(selected_kernels)
            offset_counter = self._get_offset_counter_at(current_offset)
            if not offset_counter:
                winning_kernel = self.longest_kernel
                if selected_kernels:
                    winning_kernel = selected_kernels[-1]
            else:
                for kernel in kernels:
                    if (
                        self.maximum_run_length and
                        1 < len(kernels) and
                        self.maximum_run_length <= len(selected_kernels)
                    ):
                        last_n_kernels = selected_kernels[-self.maximum_run_length:]
                        if len(set(last_n_kernels)) == 1:
                            if kernel == last_n_kernels[-1]:
                                continue
                    initial_score = kernel(offset_counter)
                    lookahead_score = self._get_lookahead_score(
                        current_offset,
                        kernel,
                        kernels,
                        )
                    score = initial_score + lookahead_score
                    kernel_score = self.KernelScore(
                        kernel=kernel,
                        score=score,
                        )
                    kernel_scores.append(kernel_score)
                kernel_scores.sort(key=lambda kernel_score: kernel_score.score)
                winning_kernel = kernel_scores[-1].kernel
            selected_kernels.append(winning_kernel)
            current_offset += winning_kernel.duration
        selected_meters = (self.kernels[_] for _ in selected_kernels)
        selected_meters = abjad.MeterList(selected_meters)
        return selected_meters

    ### PRIVATE METHODS ###

    def _get_kernels(self, selected_kernels):
        return tuple(self.kernels)

    def _get_lookahead_score(self, current_offset, kernel, kernels):
        lookahead_scores = []
        lookahead_offset = current_offset + kernel.duration
        lookahead_offset_counter = self._get_offset_counter_at(
            lookahead_offset)
        for lookahead_kernel in kernels:
            lookahead_scores.append(
                lookahead_kernel(lookahead_offset_counter)
                )
        lookahead_score = sum(lookahead_scores)  # / len(lookahead_scores)
        return lookahead_score

    def _get_offset_counter_at(self, start_offset):
        if start_offset in self.cached_offset_counters:
            return self.cached_offset_counters[start_offset]
        offset_counter = {}
        stop_offset = start_offset + self.longest_kernel.duration
        index = bisect.bisect_left(self.ordered_offsets, start_offset)
        if index == len(self.ordered_offsets):
            return offset_counter
        offset = self.ordered_offsets[index]
        while offset <= stop_offset:
            count = self.offset_counter[offset]
            offset_counter[offset - start_offset] = count
            index += 1
            if index == len(self.ordered_offsets):
                break
            offset = self.ordered_offsets[index]
        self.cached_offset_counters[start_offset] = offset_counter
        return offset_counter

    ### PUBLIC PROPERTIES ###

    @property
    def cached_offset_counters(self):
        """
        Gets cached offset counters

        Returns dictionary.
        """
        return self._cached_offset_counters

    @property
    def kernel_denominator(self):
        """
        Gets kernel denominator.

        Returns duration.
        """
        return self._kernel_denominator

    @property
    def kernels(self):
        """
        Gets kernels-to-meter dictionary.

        Returns dictionary.
        """
        return self._kernels

    @property
    def longest_kernel(self):
        """
        Gets longest kernel.

        Returns kernel.
        """
        return self._longest_kernel

    @property
    def maximum_run_length(self):
        """
        Gets maximum meter repetitions.

        Returns integer or none.
        """
        return self._maximum_run_length

    @property
    def meters(self):
        """
        Gets meters.

        Returns meters.
        """
        return self._meters

    @property
    def offset_counter(self):
        """
        Gets offset counter.

        Returns offset counter.
        """
        return self._offset_counter

    @property
    def ordered_offsets(self):
        """
        Gets ordered offsets.

        Returns offsets.
        """
        return self._ordered_offsets


class _MeterManager(system.AbjadObject):
    """
    Meter manager.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### PUBLIC METHODS ###

    @staticmethod
    def get_offsets_at_depth(depth, offset_inventory):
        """
        Gets offsets at ``depth`` in ``offset_inventory``.
        """
        import abjad
        if depth < len(offset_inventory):
            return offset_inventory[depth]
        while len(offset_inventory) <= depth:
            new_offsets = []
            old_offsets = offset_inventory[-1]
            for first, second in abjad.sequence(old_offsets).nwise():
                new_offsets.append(first)
                difference = second - first
                half = (first + second) / 2
                if abjad.Duration(1, 8) < difference:
                    new_offsets.append(half)
                else:
                    one_quarter = (first + half) / 2
                    three_quarters = (half + second) / 2
                    new_offsets.append(one_quarter)
                    new_offsets.append(half)
                    new_offsets.append(three_quarters)
            new_offsets.append(old_offsets[-1])
            offset_inventory.append(tuple(new_offsets))
        return offset_inventory[depth]

    @staticmethod
    def is_acceptable_logical_tie(
        logical_tie_duration=None,
        logical_tie_starts_in_offsets=None,
        logical_tie_stops_in_offsets=None,
        maximum_dot_count=None,
    ):
        """
        Is true if logical tie is acceptable.
        """
        #print '\tTESTING ACCEPTABILITY'
        if not logical_tie_duration.is_assignable:
            return False
        if (
            maximum_dot_count is not None and
            maximum_dot_count < logical_tie_duration.dot_count
        ):
            return False
        if (
            not logical_tie_starts_in_offsets and
            not logical_tie_stops_in_offsets
        ):
            return False
        return True

    @staticmethod
    def is_boundary_crossing_logical_tie(
        boundary_depth=None,
        boundary_offsets=None,
        logical_tie_start_offset=None,
        logical_tie_stop_offset=None,
    ):
        """
        Is true if logical tie crosses meter boundaries.
        """
        #print '\tTESTING BOUNDARY CROSSINGS'
        if boundary_depth is None:
            return False
        if not any(
            logical_tie_start_offset < x < logical_tie_stop_offset
            for x in boundary_offsets
        ):
            return False
        if (
            logical_tie_start_offset in boundary_offsets and
            logical_tie_stop_offset in boundary_offsets
        ):
            return False
        return True

    @staticmethod
    def iterate_rewrite_inputs(argument):
        r"""
        Iterates topmost masked logical ties, rest groups and containers
        in ``argument``, masked by ``argument``.

        >>> string = "abj: ! 2/4 c'4 d'4 ~ !"
        >>> string += "! 4/4 d'8. r16 r8. e'16 ~ "
        >>> string += "2/3 { e'8 ~ e'8 f'8 ~ } f'4 ~ !"
        >>> string += "! 4/4 f'8 g'8 ~ g'4 a'4 ~ a'8 b'8 ~ !"
        >>> string += "! 2/4 b'4 c''4 !"
        >>> string = string.replace('!', '|')
        >>> staff = abjad.Staff(string)

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                {   % measure
                    \time 2/4
                    c'4
                    d'4
                    ~
                }   % measure
                {   % measure
                    \time 4/4
                    d'8.
                    r16
                    r8.
                    e'16
                    ~
                    \times 2/3 {
                        e'8
                        ~
                        e'8
                        f'8
                        ~
                    }
                    f'4
                    ~
                }   % measure
                {   % measure
                    f'8
                    g'8
                    ~
                    g'4
                    a'4
                    ~
                    a'8
                    b'8
                    ~
                }   % measure
                {   % measure
                    \time 2/4
                    b'4
                    c''4
                }   % measure
            }

        >>> for x in abjad.meter._MeterManager.iterate_rewrite_inputs(
        ...     staff[0]): x
        ...
        LogicalTie([Note("c'4")])
        LogicalTie([Note("d'4")])

        >>> for x in abjad.meter._MeterManager.iterate_rewrite_inputs(
        ...     staff[1]): x
        ...
        LogicalTie([Note("d'8.")])
        LogicalTie([Rest('r16'), Rest('r8.')])
        LogicalTie([Note("e'16")])
        Tuplet(Multiplier(2, 3), "e'8 ~ e'8 f'8 ~")
        LogicalTie([Note("f'4")])

        >>> for x in abjad.meter._MeterManager.iterate_rewrite_inputs(
        ...     staff[2]): x
        ...
        LogicalTie([Note("f'8")])
        LogicalTie([Note("g'8"), Note("g'4")])
        LogicalTie([Note("a'4"), Note("a'8")])
        LogicalTie([Note("b'8")])

        >>> for x in abjad.meter._MeterManager.iterate_rewrite_inputs(
        ...     staff[3]): x
        ...
        LogicalTie([Note("b'4")])
        LogicalTie([Note("c''4")])

        Returns generator.
        """
        import abjad
        last_tie = None
        current_leaf_group = None
        current_leaf_group_is_silent = False
        for x in argument:
            if isinstance(x, (abjad.Note, abjad.Chord)):
                this_tie = x._get_spanners(abjad.Tie) or None
                if current_leaf_group is None:
                    current_leaf_group = []
                elif (
                    current_leaf_group_is_silent or
                    this_tie is None or
                    last_tie != this_tie
                ):
                    yield abjad.LogicalTie(current_leaf_group)
                    current_leaf_group = []
                current_leaf_group_is_silent = False
                current_leaf_group.append(x)
                last_tie = this_tie
            elif isinstance(x, (abjad.Rest, abjad.Skip)):
                if current_leaf_group is None:
                    current_leaf_group = []
                elif not current_leaf_group_is_silent:
                    yield abjad.LogicalTie(current_leaf_group)
                    current_leaf_group = []
                current_leaf_group_is_silent = True
                current_leaf_group.append(x)
                last_tie = None
            elif isinstance(x, abjad.Container):
                if current_leaf_group is not None:
                    yield abjad.LogicalTie(current_leaf_group)
                    current_leaf_group = None
                    last_tie = None
                yield x

            else:
                message = 'unhandled component: {!r}.'
                message = message.format(x)
                raise Exception(message)
        if current_leaf_group is not None:
            yield abjad.LogicalTie(current_leaf_group)



import collections


def cumulative_products(argument):
    """
    Gets cumulative products of ``argument``.

    ..  container:: example

        >>> abjad.mathtools.cumulative_products([1, 2, 3, 4, 5, 6, 7, 8])
        [1, 2, 6, 24, 120, 720, 5040, 40320]

        >>> abjad.mathtools.cumulative_products([1, -2, 3, -4, 5, -6, 7, -8])
        [1, -2, -6, 24, 120, -720, -5040, 40320]

    Raises exception when ``argument`` is not iterable.

    Returns new object of ``argument`` type.
    """
    if not isinstance(argument, collections.Iterable):
        raise TypeError(argument)
    if len(argument) == 0:
        return type(argument)([])
    result = [argument[0]]
    for element in argument[1:]:
        result.append(result[-1] * element)
    return type(argument)(result)



def partition_integer_into_canonic_parts(n, decrease_parts_monotonically=True):
    """
    Partitions integer ``n`` into canonic parts.

    ..  container:: example

        Returns all parts positive on positive ``n``:

        >>> for n in range(1, 11):
        ...     print(n, abjad.mathtools.partition_integer_into_canonic_parts(n))
        ...
        1 (1,)
        2 (2,)
        3 (3,)
        4 (4,)
        5 (4, 1)
        6 (6,)
        7 (7,)
        8 (8,)
        9 (8, 1)
        10 (8, 2)

    ..  container:: example

        Returns all parts negative on negative ``n``:

        >>> for n in reversed(range(-20, -10)):
        ...     print(n, abjad.mathtools.partition_integer_into_canonic_parts(n))
        ...
        -11 (-8, -3)
        -12 (-12,)
        -13 (-12, -1)
        -14 (-14,)
        -15 (-15,)
        -16 (-16,)
        -17 (-16, -1)
        -18 (-16, -2)
        -19 (-16, -3)
        -20 (-16, -4)

    ..  container:: example

        Returns parts that increase monotonically:

        >>> for n in range(11, 21):
        ...     print(n, abjad.mathtools.partition_integer_into_canonic_parts(n,
        ...         decrease_parts_monotonically=False))
        ...
        11 (3, 8)
        12 (12,)
        13 (1, 12)
        14 (14,)
        15 (15,)
        16 (16,)
        17 (1, 16)
        18 (2, 16)
        19 (3, 16)
        20 (4, 16)

    Returns tuple with parts that decrease monotonically.

    Returns tuple of integers.
    """
    from abjad import mathtools
    assert isinstance(n, int), repr(n)
    assert isinstance(decrease_parts_monotonically, bool)
    if n == 0:
        return (0,)
    result = []
    previous_empty = True
    binary_n = mathtools.integer_to_binary_string(abs(n))
    binary_length = len(binary_n)
    for i, character in enumerate(binary_n):
        if character == '1':
            place_value = 2 ** (binary_length - i - 1)
            if previous_empty:
                result.append(place_value)
            else:
                result[-1] += place_value
            previous_empty = False
        else:
            previous_empty = True
    sign_n = mathtools.sign(n)
    if mathtools.sign(n) == -1:
        result = [sign_n * _ for _ in result]
    if decrease_parts_monotonically:
        return tuple(result)
    else:
        return tuple(reversed(result))
